#!python
import os
import json
import logging
import platform
import pandas as pd

# visualization
import panel as pn
import bokeh.server.views.ws

# modules
import pydiaid
import pydiaid.loader
import pydiaid.main
import pydiaid.graphs
import pydiaid.method_optimizer
import pydiaid.method_evaluation

import warnings
warnings.filterwarnings("ignore")

# paths
BASE_PATH = os.path.dirname(__file__)
IMG_PATH = os.path.join(BASE_PATH, "img")
STYLE_PATH = os.path.join(BASE_PATH, "style")
DOCS_PATH = os.path.join(BASE_PATH, "docs")
DEFAULT_FILE = os.path.join(
    BASE_PATH,
    'lib',
    "default_parameters.json"
)

with open(DEFAULT_FILE, "r") as infile:
    method_conf = json.load(infile)

if platform.system() == 'Windows':
    method_path_placeholder = 'D:\pydiaid\static\DIAParameterspy3TC.txt'
    library_path_placeholder = 'D:\pydiaid\static\AlphaPept_results.csv'
    save_folder_placeholder = 'D:\pydiaid\static'
else:
    method_path_placeholder = '/Users/pydiaid/static/DIAParameterspy3TC.txt'
    library_path_placeholder = '/Users/pydiaid/static/AlphaPept_results.csv'
    save_folder_placeholder = '/Users/pydiaid/static'


class BaseWidget(object):
    # TODO: docstring
    def __init__(self, name):
        self.name = name
        self.__update_event = pn.widgets.IntInput(value=0)
        self.depends = pn.depends(self.__update_event.param.value)
        self.active_depends = pn.depends(
            self.__update_event.param.value,
            watch=True
        )

    def trigger_dependancy(self):
        self.__update_event.value += 1


class HeaderWidget(object):
    """This class creates a layout for the header of the dashboard with the name of the tool and all links to the MPI website, the MPI Mann Lab page and the GitHub repo.

    Parameters
    ----------
    title : str
        The name of the tool.
    img_folder_path : str
        The path to the folder with all necessary image files.
    github_url : str
        The path to the GitHub repository of the project.

    Attributes
    ----------
    header_title : pn.pane.Markdown
        A Panel Markdown pane that returns the title of the tool.
    mpi_biochem_logo : pn.pane.PNG
        A Panel PNG pane that embeds a png image file of the MPI Biochemisty logo and makes the image clickable with the link to the official website of the department.
    mpi_logo : pn.pane.JPG
        A Panel JPG pane that embeds a jpg image file of the MPI Biochemisty logo and makes the image clickable with the link to the official website.
    github_logo : pn.pane.PNG
        A Panel PNG pane that embeds a png image file of the GitHub logo and makes the image clickable with the link to the GitHub repository of the project.
    """
    def __init__(
        self,
        title,
        img_folder_path,
        github_url
    ):
        self.header_title = pn.pane.Markdown(
            f'# {title}',
            sizing_mode='stretch_width',
        )
        self.biochem_logo_path = os.path.join(
            img_folder_path,
            "mpi_logo.png"
        )
        self.mpi_logo_path = os.path.join(
            img_folder_path,
            "max-planck-gesellschaft.jpg"
        )
        self.github_logo_path = os.path.join(
            img_folder_path,
            "github.png"
        )
        self.mpi_biochem_logo = pn.pane.PNG(
            self.biochem_logo_path,
            link_url='https://www.biochem.mpg.de/mann',
            width=60,
            height=60,
            align='start'
        )
        self.mpi_logo = pn.pane.JPG(
            self.mpi_logo_path,
            link_url='https://www.biochem.mpg.de/en',
            height=62,
            embed=True,
            width=62,
            margin=(5, 0, 0, 5),
            css_classes=['opt']
        )
        self.github_logo = pn.pane.PNG(
            self.github_logo_path,
            link_url=github_url,
            height=70,
            align='end'
        )

    def create(self):
        layout = pn.Row(
            self.mpi_biochem_logo,
            self.mpi_logo,
            self.header_title,
            self.github_logo,
            height=73,
            sizing_mode='stretch_width'
        )
        return layout


class MainWidget(object):
    """This class creates a layout for the main part of the dashboard with the description of the tool and a button to download the manual for the project's GUI.

    Parameters
    ----------
    description : str
        The short description of the tool.
    manual_path : str
        The path to the GUI manual.

    Attributes
    ----------
    project_description : pn.pane.Markdown
        A Panel Markdown pane that shows the description of the project.
    manual : pn.widgets.FileDownload
        A Panel FileDownload widget that allows to download the GUI manual of the tool.
    """

    def __init__(
        self,
        description,
        manual_path
    ):
        self.project_description = pn.pane.Markdown(
            description,
            margin=(10, 0, 10, 0),
            css_classes=['main-part'],
            align='start',
            width=460
        )
        self.manual = pn.widgets.FileDownload(
            file=manual_path,
            label='Download Manual',
            button_type='default',
            align='center',
            auto=True,
            height=31,
            width=200,
            margin=(0, 20, 0, 0)
        )

    def create(self):
        layout = pn.Row(
            self.project_description,
            pn.layout.HSpacer(width=500),
            self.manual,
            background='#eaeaea',
            align='center',
            sizing_mode='stretch_width',
            height=190,
            margin=(10, 8, 10, 8),
            css_classes=['background']
        )
        return layout


class LoadLibraryCard(BaseWidget):
    # TODO: docstring
    def __init__(self, description):
        super().__init__(name="Data")
        self.library = None
        self.layout = None
        self.path_library = pn.widgets.TextInput(
            name='Specify the path to the library:',
            placeholder=library_path_placeholder,
            # value=method_conf["input"]["library_name"],
            width=900,
            sizing_mode='stretch_width',
            margin=(15, 15, 0, 15)
        )
        self.path_save_folder = pn.widgets.TextInput(
            name='Save the output to the following folder:',
            # value=method_conf["input"]["save_at"],
            placeholder=save_folder_placeholder,
            width=900,
            sizing_mode='stretch_width',
            margin=(15, 15, 0, 15)
        )
        self.ptm = pn.widgets.TextInput(
            name='Specify the PTM:',
            value=method_conf["input"]["PTM"],
            placeholder='Phospho',
            width=680,
            sizing_mode='stretch_width',
            margin=(15, 15, 15, 15)
        )
        self.analysis_software = pn.widgets.Select(
            name='Analysis software',
            value=method_conf["input"]["analysis_software"],
            options=['AlphaPept', 'MaxQuant', 'FragPipe', 'Spectronaut single-run', 'Spectronaut library'],
            width=200,
            margin=(15, 15, 15, 15)
        )
        # PLOTS
        self.plot_mz = pn.widgets.EditableRangeSlider(
            name='Plot m/z-range [Da]',
            start=100,
            end=1700,
            value=tuple(method_conf['graphs']['plot_mz']),
            step=50,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.plot_im = pn.widgets.EditableRangeSlider(
            name='Plot ion mobility range [1/K0]',
            start=0.6,
            end=1.6,
            value=tuple(method_conf['graphs']['plot_IM']),
            step=0.1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.numbins = pn.widgets.IntInput(
            name='Number of bins',
            start=1,
            end=300,
            value=method_conf['graphs']['numbins'],
            step=1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.window_transparency = pn.widgets.FloatInput(
            name='Transparency',
            start=0.1,
            end=1,
            value=method_conf['graphs']['window_transparency'],
            step=0.1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.window_frame_color = pn.widgets.Select(
            name='Frame color',
            value=method_conf['graphs']['window_frame_color'],
            options=['black', 'grey'],
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.window_color = pn.widgets.Select(
            name='Color',
            value=method_conf['graphs']['window_color'],
            options=['yellow', 'green', 'black', 'grey', 'white'],
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.load_library_descr = pn.pane.Markdown(
            description,
            margin=(2, 0, 2, 0),
            # css_classes=['main-part'],
            align='center',
            # width=460
        )
        # UPLOAD DATA
        self.upload_button = pn.widgets.Button(
            name='Upload library',
            button_type='primary',
            height=31,
            width=250,
            align='center',
            margin=(0, 0, 0, 0)
        )
        self.upload_progress = pn.indicators.Progress(
            active=False,
            bar_color='light',
            width=250,
            align='center',
            margin=(-10, 0, 30, 0)
        )
        self.import_error = pn.pane.Alert(
            alert_type="danger",
            sizing_mode='stretch_width',
            # object='test warning message',
            margin=(30, 15, 5, 15),
        )

    def create(self):
        self.layout = pn.Card(
            self.load_library_descr,
            pn.Row(
                pn.Column(
                    self.path_library,
                    self.path_save_folder,
                    pn.Row(
                        self.ptm,
                        self.analysis_software,
                    ),
                    margin=(10, 30, 10, 10),
                ),
                pn.Spacer(sizing_mode='stretch_width'),
                pn.Column(
                    self.upload_button,
                    self.upload_progress,
                    # self.import_error,
                    align='center',
                    margin=(0, 130, 0, 0),
                )
            ),
            pn.Column(
                pn.WidgetBox(
                    pn.Row(
                        self.plot_mz,
                        self.plot_im,
                        sizing_mode='stretch_width',
                    ),
                    pn.Row(
                        self.numbins,
                        self.window_transparency,
                        sizing_mode='stretch_width'
                    ),
                    pn.Row(
                        self.window_frame_color,
                        self.window_color,
                        sizing_mode='stretch_width'
                    ),
                    sizing_mode='stretch_width',
                    margin=(10, 10, 30, 10),
                    height=220
                ),
                margin=(10, 30, 10, 10),
            ),
            pn.layout.Divider(
                sizing_mode='stretch_width',
                margin=(0, 10, -20, 10),
            ),
            pn.Row(
                None, None, None
            ),
            title='Load Library',
            collapsed=False,
            collapsible=True,
            header_background='#eaeaea',
            background ='white',
            header_color='#333',
            align='center',
            sizing_mode='stretch_width',
            margin=(5, 8, 10, 8),
            css_classes=['background']
        )

        dependances = {
            self.path_library: [self.update_parameters, 'value'],
            self.path_save_folder: [self.update_parameters, 'value'],
            self.ptm: [self.update_parameters, 'value'],
            self.analysis_software: [self.update_parameters, 'value'],
            self.plot_mz: [self.update_parameters_plotting, 'value'],
            self.plot_im: [self.update_parameters_plotting, 'value'],
            self.numbins: [self.update_parameters_plotting, 'value'],
            self.window_transparency: [self.update_parameters_plotting, 'value'],
            self.window_frame_color: [self.update_parameters_plotting, 'value'],
            self.window_color: [self.update_parameters_plotting, 'value'],
            self.upload_button: [self.upload_data, 'clicks'],

        }
        for k in dependances.keys():
            k.param.watch(
                dependances[k][0],
                dependances[k][1],
                onlychanged=True
            )

        return self.layout


    def update_parameters(self, event):
        global method_conf
        convertion_dict = {
            self.path_library.name: "library_name",
            self.path_save_folder.name: "save_at",
            self.ptm.name: "PTM",
            self.analysis_software.name: "analysis_software",
        }
        method_conf['input'][convertion_dict[event.obj.name]] = event.new

    def update_parameters_plotting(self, event):
        global method_conf
        convertion_dict = {
            self.plot_mz.name: "plot_mz",
            self.plot_im.name: "plot_IM",
            self.numbins.name: "numbins",
            self.window_frame_color.name: "window_frame_color",
            self.window_color.name: "window_color",
            self.window_transparency.name: "window_transparency",
        }
        method_conf['graphs'][convertion_dict[event.obj.name]] = event.new

    def upload_data(self, *args):
        self.upload_progress.active = True
        self.library = pydiaid.loader.load_library(
            self.path_library.value,
            method_conf["input"]["analysis_software"],
            method_conf["input"]["PTM"]
        )
        self.upload_progress.active = False
        folder_paths = [
            self.path_save_folder.value,
            os.path.join(
                self.path_save_folder.value,
                'input_library'
            ),
            os.path.join(
                self.path_save_folder.value,
                'final_method'
            ),
        ]
        pydiaid.main.create_folder(folder_paths)
        self.xi, self.yi, self.zi = pydiaid.graphs.kernel_density_calculation(
            self.library,
            method_conf["graphs"]["numbins"]
        )
        self.layout[4][0] = pn.pane.Matplotlib(
            pydiaid.graphs.plot_precursor_distribution_as_histogram(
                self.library,
                method_conf["graphs"],
                os.path.join(
                    self.path_save_folder.value,
                    'input_library',
                    'Histogram_precursor_distribution_in_library.png'
                ),
                gui=True
            ),
            tight=True
        )
        self.layout[4][1] = pn.pane.Matplotlib(
            pydiaid.graphs.plot_density(
                self.xi,
                self.yi,
                self.zi,
                method_conf["graphs"],
                os.path.join(
                    self.path_save_folder.value,
                    'input_library',
                    'Kernel_density_distribution_library.png'
                ),
                gui=True
            ),
            tight=True
        )
        dict_charge_of_precursor = pydiaid.method_evaluation.calculate_percentage_multiple_charged_precursors(self.library)
        mult_charged_precursor_info = pd.DataFrame(
            {
                "charge state of precursors": list(dict_charge_of_precursor.keys()),
                "ratio [%]": list(dict_charge_of_precursor.values())
            }
        )
        mult_charged_precursor_info.to_csv(
            os.path.join(
                self.path_save_folder.value,
                'input_library',
                'percentage_of_multiple_charged_precursors.csv'
            ),
            index=False
        )
        self.layout[4][2] = pn.Column(
            pn.pane.Markdown(
                '### Percentage of multiple charged precursors',
                 align='center'
            ),
            pn.widgets.DataFrame(
                mult_charged_precursor_info,
                autosize_mode='fit_viewport',
                auto_edit=False
            ),
            margin=(0, 50),
            sizing_mode='stretch_width',
            align='center'
        )
        self.trigger_dependancy()
        self.upload_progress.active = False


class SpecifyParametersCard(BaseWidget):
    # TODO: docstring
    def __init__(self, data, description):
        super().__init__(name="Parameters")
        self.data = data
        self.mz = pn.widgets.EditableRangeSlider(
            name='m/z-range [Da]',
            start=100,
            end=1700,
            value=tuple(method_conf['method_parameters']['mz']),
            step=50,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.ion_mobility = pn.widgets.EditableRangeSlider(
            name='Ion mobility range [1/K0]',
            start=0.6,
            end=1.6,
            value=tuple(method_conf['method_parameters']['ion_mobility']),
            step=0.05,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.num_dia_pasef_scans = pn.widgets.IntInput(
            name='Number of dia-PASEF scans',
            start=1,
            end=40,
            value=method_conf['method_parameters']['num_dia_pasef_scans'],
            step=1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.im_steps = pn.widgets.IntInput(
            name='Number of ion mobility windows / dia-PASEF scan',
            start=1,
            end=10,
            value=method_conf['method_parameters']['im_steps'],
            step=1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.overlap = pn.widgets.IntInput(
            name='Isolation window overlap [Da]',
            start=0,
            end=10,
            value=method_conf['method_parameters']['overlap'],
            step=1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.shift_of_final_method = pn.widgets.FloatInput(
            name='Shift of the final acquisition scheme (in IM dimension) [1/K0]',
            start=-0.5,
            end=1.0,
            value=method_conf['method_parameters']['shift_of_final_method'],
            step=0.01,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.spec_param_table = pn.widgets.DataFrame(
            autosize_mode='fit_viewport',
            # margin=(0, 0, 0, 100),
            align='center',
            auto_edit=False,
        )
        self.specify_parameter_descr = pn.pane.Markdown(
            description,
            margin=(2, 0, 2, 0),
            # css_classes=['main-part'],
            align='center',
            # width=460
        )
        self.calculate_button = pn.widgets.Button(
            name='Calculate',
            button_type='primary',
            height=31,
            width=250,
            align='center',
            margin=(0, 0, 0, 0)
        )


    def create(self):
        self.layout = pn.Card(
            self.specify_parameter_descr,
            pn.Row(
                pn.Column(
                    pn.WidgetBox(
                        pn.Row(
                            self.mz,
                            self.ion_mobility,
                            sizing_mode='stretch_width',
                        ),
                        pn.Row(
                            self.num_dia_pasef_scans,
                            self.im_steps,
                            sizing_mode='stretch_width'
                        ),
                        pn.Row(
                            self.overlap,
                            self.shift_of_final_method,
                            sizing_mode='stretch_width'
                        ),
                        sizing_mode='stretch_width',
                        margin=(20, 10, 30, 10),
                        height=220
                    ),
                    margin=(10, 30, 10, 10),
                ),
                pn.Spacer(sizing_mode='stretch_width'),
                pn.Column(
                    self.calculate_button,
                    align='center',
                    margin=(0, 130, 0, 0),
                )
            ),
            pn.layout.Divider(
                sizing_mode='stretch_width',
                margin=(-20, 10, -20, 10),
            ),
            pn.Row(
                None
            ),
            title='Specify Method Parameters',
            collapsed=True,
            collapsible=True,
            header_background='#eaeaea',
            background ='white',
            header_color='#333',
            align='center',
            sizing_mode='stretch_width',
            # height=470,
            margin=(5, 8, 10, 8),
            css_classes=['background']
        )
        dependances = {
            self.mz: [self.update_parameters, 'value'],
            self.ion_mobility: [self.update_parameters, 'value'],
            self.num_dia_pasef_scans: [self.update_parameters, 'value'],
            self.im_steps: [self.update_parameters, 'value'],
            self.overlap: [self.update_parameters, 'value'],
            self.shift_of_final_method: [self.update_parameters, 'value'],
            self.calculate_button: [self.run_calculation, 'clicks'],
        }
        for k in dependances.keys():
            k.param.watch(
                dependances[k][0],
                dependances[k][1],
                onlychanged=True
            )
        return self.layout

    def update_parameters(self, event):
        global method_conf
        convertion_dict = {
            self.mz.name: "mz",
            self.ion_mobility.name: "ion_mobility",
            self.num_dia_pasef_scans.name: "num_dia_pasef_scans",
            self.im_steps.name: "im_steps",
            self.overlap.name: "overlap",
            self.shift_of_final_method.name: "shift_of_final_method",
        }
        if isinstance(event.new, tuple):
            method_conf['method_parameters'][convertion_dict[event.obj.name]] = list(event.new)
        else:
            method_conf['method_parameters'][convertion_dict[event.obj.name]] = event.new

    def run_calculation(self, *args):
        dict_precursors_within_scan_area = pydiaid.method_evaluation.calculate_precursor_within_scan_area(
            self.data.library,
            method_conf['method_parameters']["mz"],
            method_conf['method_parameters']["ion_mobility"]
        )

        df = pd.DataFrame({
            "precursors within the scan area [%]": list(dict_precursors_within_scan_area.values())
        })

        df.to_csv(
            os.path.join(
                method_conf["input"]["save_at"],
                'final_method',
                'precursors_within_scan_area.csv'
            ),
            index=False
        )

        self.spec_param_table.value = df
        self.layout[3][0] = self.spec_param_table

class OptimizationCard(BaseWidget):
    # TODO: docstring
    def __init__(self, data, description):
        super().__init__(name="Optimize Method")
        self.data = data
        self.opt_result_x = [0, 0, 0, 0]
        self.n_calls = pn.widgets.IntInput(
            name='Number of iterative optimization steps',
            start=1,
            end=200,
            value=method_conf['optimizer']['n_calls'],
            step=1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        #self.n_start = pn.widgets.IntInput(
        #    name='Number of starting points',
        #    start=1,
        #    end=20,
        #    value=method_conf['optimizer']['n_start'],
        #    step=1,
        #    margin=(15, 15, 0, 15),
        #    width=430,
        #)
        self.initial_points = pn.widgets.IntInput(
            name='Number of starting points',
            start=1,
            end=20,
            value=method_conf['optimizer']['initial_points'],
            step=1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.YA1 = pn.widgets.EditableRangeSlider(
            name='A1 range',
            start=0.2,
            end=2.0,
            value=tuple(method_conf['optimizer']['YA1']),
            step=0.1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.YA2 = pn.widgets.EditableRangeSlider(
            name='A2 range',
            start=0.2,
            end=2.0,
            value=tuple(method_conf['optimizer']['YA2']),
            step=0.05,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.YB1 = pn.widgets.EditableRangeSlider(
            name='B1 range',
            start=0.2,
            end=2.0,
            value=tuple(method_conf['optimizer']['YB1']),
            step=0.1,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.YB2 = pn.widgets.EditableRangeSlider(
            name='B2 range',
            start=0.2,
            end=2.0,
            value=tuple(method_conf['optimizer']['YB2']),
            step=0.05,
            margin=(15, 15, 0, 15),
            width=430,
        )
        self.evaluation_parameter = pn.widgets.Select(
            name='Evaluation parameter',
            value=method_conf['optimizer']['evaluation_parameter'],
            options=[
                "No. of covered proteins",
                'No. of covered precursors',
                "No. of covered, doubly charged precursors",
                "No. of covered, triply charged precursors",
                "No. of covered, quadruply charged precursors"
            ],
            width=430,
            margin=(15, 15, 0, 15)
        )
        self.optimize_button = pn.widgets.Button(
            name='Optimize',
            button_type='primary',
            height=31,
            width=250,
            align='center',
            margin=(0, 0, 0, 0)
        )
        self.optimize_spinner = pn.indicators.LoadingSpinner(
            value=False,
            bgcolor='light',
            color='secondary',
            align='center',
            margin=(0, 0, 0, 20),
            width=35,
            height=35
        )
        self.scan_area_A1_A2_B1_B2_only_used_for_specific_diaPASEF = pn.widgets.LiteralInput(
            name='Scan area A1/A2/B1/B2',
            # value=method_conf['method_parameters']['scan_area_A1_A2_B1_B2_only_used_for_specific_diaPASEF'],
            value=[0,0,0,0],
            type=list,
            margin=(15, 15, 0, 15),
            width=900
        )
        self.optimization_descr = pn.pane.Markdown(
            description,
            margin=(2, 0, 2, 0),
            # css_classes=['main-part'],
            align='center',
            # width=460
        )

    def create(self):
        self.layout = pn.Card(
            self.optimization_descr,
            pn.Row(
                pn.Column(
                    pn.WidgetBox(
                        pn.Row(
                            self.n_calls,
                            #self.n_start,
                            self.initial_points,
                            sizing_mode='stretch_width',
                        ),
                        pn.Row(
                            #self.initial_points,
                            self.evaluation_parameter,
                            sizing_mode='stretch_width'
                        ),
                        pn.Row(
                            self.YA1,
                            self.YA2,
                            sizing_mode='stretch_width'
                        ),
                        pn.Row(
                            self.YB1,
                            self.YB2,
                            sizing_mode='stretch_width'
                        ),
                        sizing_mode='stretch_width',
                        margin=(20, 10, 30, 10),
                        height=270
                    ),
                    margin=(10, 30, 10, 10),
                ),
                pn.Spacer(sizing_mode='stretch_width'),
                pn.Row(
                    self.optimize_button,
                    self.optimize_spinner,
                    align='start',
                    margin=(150, 10, 0, 0),
                )
            ),
            pn.layout.Divider(
                sizing_mode='stretch_width',
                margin=(-20, 10, -20, 10),
            ),
            pn.Column(
                None,
                pn.Row(
                    None,
                    None,
                ),
                align='center',
            ),
            title='Optimization',
            collapsed=True,
            collapsible=True,
            header_background='#eaeaea',
            background ='white',
            header_color='#333',
            align='center',
            sizing_mode='stretch_width',
            # height=470,
            margin=(5, 8, 10, 8),
            css_classes=['background']
        )

        dependances = {
            self.n_calls: [self.update_parameters, 'value'],
            #self.n_start: [self.update_parameters, 'value'],
            self.initial_points: [self.update_parameters, 'value'],
            self.evaluation_parameter: [self.update_parameters, 'value'],
            self.YA1: [self.update_parameters, 'value'],
            self.YA2: [self.update_parameters, 'value'],
            self.YB1: [self.update_parameters, 'value'],
            self.YB2: [self.update_parameters, 'value'],
            self.optimize_button: [self.run_optimization, 'clicks'],
        }
        for k in dependances.keys():
            k.param.watch(
                dependances[k][0],
                dependances[k][1],
                onlychanged=True
            )
        return self.layout

    def update_parameters(self, event):
        global method_conf
        convertion_dict = {
            self.n_calls.name: "n_calls",
            #self.n_start.name: "n_start",
            self.initial_points: "initial_points",
            self.evaluation_parameter.name: "evaluation_parameter",
            self.YA1.name: "YA1",
            self.YA2.name: "YA2",
            self.YB1.name: "YB1",
            self.YB2.name: "YB2",
        }
        if isinstance(event.new, tuple):
            method_conf['optimizer'][convertion_dict[event.obj.name]] = list(event.new)
        else:
            method_conf['optimizer'][convertion_dict[event.obj.name]] = event.new

    def run_optimization(self, *args):
        self.optimize_spinner.value = True

        self.folder_path = [
            os.path.join(
                method_conf["input"]["save_at"],
                'optimization_plots'
            )
        ]
        pydiaid.main.create_folder(self.folder_path)

        self.opt_result = pydiaid.method_optimizer.optimization(
            self.data.library,
            method_conf["method_parameters"],
            self.data.xi,
            self.data.yi,
            self.data.zi,
            method_conf,
            method_conf["optimizer"]
        )

        self.scan_area_A1_A2_B1_B2_only_used_for_specific_diaPASEF.value = self.opt_result

        self.filenames_plots =  pydiaid.loader.get_file_names_from_directory(
            self.folder_path[0],
            'png'
        )

        self.player = pn.widgets.Player(
            start=0,
            end=len(self.filenames_plots)-1,
            value=0,
            loop_policy='loop',
            align='center',
            margin=(0, 0, 50, 0)
        )

        opt_plot_df = pydiaid.loader.create_opt_plot_df(
            self.filenames_plots[self.player.value]
        )

        self.kde_plot = pn.pane.PNG(
            object=os.path.join(
                self.folder_path[0],
                self.filenames_plots[self.player.value]
            ),
            height=345,
            width=460,
            align='center',
        )
        self.kde_plot_table = pn.widgets.DataFrame(
            opt_plot_df,
            autosize_mode='fit_viewport',
            margin=(0, 0, 0, 100),
            align='center',
            auto_edit=False
        )

        self.layout[3][0] = self.player
        self.layout[3][1][0] = self.kde_plot
        self.layout[3][1][1] = self.kde_plot_table

        self.player.param.watch(
            self.update_kde_plot_df,
            'value'
        )

        self.trigger_dependancy()
        self.optimize_spinner.value = False

    def update_kde_plot_df(self, event):
        self.kde_plot_table.value = pydiaid.loader.create_opt_plot_df(
            self.filenames_plots[event.new]
        )
        self.kde_plot.object = os.path.join(
            self.folder_path[0],
            self.filenames_plots[event.new]
        )
        self.layout[3][1][0] = self.kde_plot
        self.layout[3][1][1] = self.kde_plot_table


class CreateMethodCard(BaseWidget):
    # TODO: docstring
    def __init__(self, data, opt_widget, description):
        super().__init__(name="Method_creation")
        self.data = data
        self.opt_widget = opt_widget
        self.create_button = pn.widgets.Button(
            name='Create',
            button_type='primary',
            height=31,
            width=250,
            align='center',
            margin=(0, 0, 0, 0)
        )
        self.path_method = pn.widgets.TextInput(
            name='Specify the path to the method file:',
            placeholder=method_path_placeholder,
            # value=method_conf['input']['diaPASEF_method_only_used_for_method_evaluation'],
            width=900,
            sizing_mode='stretch_width',
            margin=(15, 15, 0, 15)
        )
        self.create_method_descr = pn.pane.Markdown(
            description,
            margin=(2, 0, 2, 0),
            # css_classes=['main-part'],
            align='center',
            # width=460
        )

    def create(self):
        self.layout = pn.Card(
            self.create_method_descr,
            pn.Row(
                pn.Column(
                    pn.WidgetBox(
                        pn.Row(
                            self.opt_widget.scan_area_A1_A2_B1_B2_only_used_for_specific_diaPASEF,
                            sizing_mode='stretch_width'
                        ),
                        sizing_mode='stretch_width',
                        margin=(20, 10, 30, 10),
                        height=90
                    ),
                    margin=(10, 30, 10, 10),
                ),
                pn.Spacer(sizing_mode='stretch_width'),
                pn.Row(
                    self.create_button,
                    align='center',
                    margin=(0, 130, 0, 0),
                )
            ),
            pn.layout.Divider(
                sizing_mode='stretch_width',
                margin=(-20, 10, -20, 10),
            ),
            pn.Row(
                None,
                None,
                align='center'
            ),
            title='Create Method',
            collapsed=True,
            collapsible=True,
            header_background='#eaeaea',
            background ='white',
            header_color='#333',
            align='center',
            sizing_mode='stretch_width',
            # height=470,
            margin=(5, 8, 10, 8),
            css_classes=['background']
        )

        dependances = {
            self.opt_widget.scan_area_A1_A2_B1_B2_only_used_for_specific_diaPASEF: [self.update_parameters, 'value'],
            self.create_button: [self.create_method, 'clicks'],
        }
        for k in dependances.keys():
            k.param.watch(
                dependances[k][0],
                dependances[k][1],
                onlychanged=True
            )
        return self.layout

    def update_parameters(self, event):
        global method_conf
        convertion_dict = {
            self.opt_widget.scan_area_A1_A2_B1_B2_only_used_for_specific_diaPASEF.name: "scan_area_A1_A2_B1_B2_only_used_for_create",
        }
        method_conf['method_parameters'][convertion_dict[event.obj.name]] = event.new

    def create_method(self, event):
        df_parameters_final = pydiaid.main.create_final_method(
            self.data.library,
            method_conf["method_parameters"],
            self.opt_widget.scan_area_A1_A2_B1_B2_only_used_for_specific_diaPASEF.value,
            method_conf
        )

        # save input_parameter as .csv
        #pd.DataFrame(
        #    {
        #    "parameters": list(method_conf["input"].keys()) +
        #        list(method_conf["method_parameters"].keys()) +
        #        list(method_conf["graphs"].keys()) +
        #        list(method_conf["optimizer"].keys()),
        #    "value": list(method_conf["input"].values()) +
        #        list(method_conf["method_parameters"].values()) +
        #        list(method_conf["graphs"].values()) +
        #        list(method_conf["optimizer"].values())
        #    }
        #).to_csv(
        #    os.path.join(
        #        method_conf["input"]["save_at"],
        #        'final_method',
        #        'input_parameters.csv'
        #    ),
        #    index=False
        #)

        # save input parameters as json file for reusing
        out_file = open(
                method_conf["input"]["save_at"]+'/input_parameters.json', "w")
        json.dump(method_conf, out_file, indent = 4)
        out_file.close()

        final_method_path = os.path.join(
            method_conf["input"]["save_at"],
            'final_method',
            'diaPASEF_method.txt'
        )
        self.path_method.value = final_method_path
        self.layout[3][0] = pn.widgets.DataFrame(
            pd.read_csv(
                final_method_path,
                header=None,
                skiprows=3,
                names=["MS Type", "Cycle Id", "Start IM", "End IM", "Start Mass", "End Mass", "CE"]
            ),
            autosize_mode='fit_viewport',
            margin=(0, 0, 0, 100),
            align='center',
            auto_edit=False,
            height=240,
        )
        self.trigger_dependancy()

class EvaluateMethodCard(object):
    # TODO: docstring
    def __init__(self, data, method_creation, description):
        self.data = data
        self.method_creation = method_creation
        self.evaluate_button = pn.widgets.Button(
            name='Evaluate',
            button_type='primary',
            height=31,
            width=250,
            align='center',
            margin=(0, 0, 0, 0)
        )
        self.evaluate_method_descr = pn.pane.Markdown(
            description,
            margin=(2, 0, 2, 0),
            # css_classes=['main-part'],
            align='center',
            # width=460
        )

    def create(self):
        self.layout = pn.Card(
            self.evaluate_method_descr,
            pn.Row(
                pn.Column(
                    self.method_creation.path_method,
                    sizing_mode='stretch_width',
                    margin=(20, 10, 30, 10),
                ),
                pn.Spacer(sizing_mode='stretch_width'),
                pn.Row(
                    self.evaluate_button,
                    align='center',
                    margin=(0, 130, 0, 0),
                )
            ),
            pn.layout.Divider(
                sizing_mode='stretch_width',
                margin=(-20, 10, -20, 10),
            ),
            pn.Row(
                None,
                None,
                align='center'
            ),
            title='Evaluate Method',
            collapsed=True,
            collapsible=True,
            header_background='#eaeaea',
            background ='white',
            header_color='#333',
            align='center',
            sizing_mode='stretch_width',
            # height=470,
            margin=(5, 8, 10, 8),
            css_classes=['background']
        )

        dependances = {
            self.method_creation.path_method: [self.update_parameters, 'value'],
            self.evaluate_button: [self.evaluate_method, 'clicks'],
        }
        for k in dependances.keys():
            k.param.watch(
                dependances[k][0],
                dependances[k][1],
                onlychanged=True
            )
        return self.layout

    def update_parameters(self, event):
        global method_conf
        convertion_dict = {
            self.method_creation.path_method.name: "diaPASEF_method_only_used_for_method_evaluation",
        }
        method_conf['input'][convertion_dict[event.obj.name]] = event.new

    def evaluate_method(self, event):
        method_eval_table = pn.widgets.DataFrame(
            autosize_mode='fit_viewport',
            # margin=(0, 0, 0, 100),
            # align='center',
            auto_edit=False,
        )

        method_eval_dir = os.path.dirname(self.method_creation.path_method.value)

        df_parameters_final = pd.read_csv(
            method_conf["input"]["diaPASEF_method_only_used_for_method_evaluation"],
            skiprows=4,
            names=["MS Type", "Cycle Id", "Start IM", "End IM", "Start Mass", "End Mass", "CE"]
        )

        pydiaid.main.final_method_information(
            df_parameters_final,
            self.data.xi,
            self.data.yi,
            self.data.zi,
            method_conf,
            method_conf["input"]["save_at"],
            self.data.library,
            method_conf["method_parameters"],
            method_conf["method_parameters"]["scan_area_A1_A2_B1_B2_only_used_for_create"]
        )

        # save parameters for method evaluation as .csv
        dict_precursors_within_mz = pydiaid.method_evaluation.calculate_precursor_within_scan_area(
            self.data.library,
            method_conf["method_parameters"]["mz"],
            method_conf["method_parameters"]["ion_mobility"]
        )
        dict_precursors_coverage = pydiaid.method_evaluation.coverage(
            df_parameters_final,
            self.data.library,
        )
        dict_evaluation_of_final_method = {
            **dict_precursors_within_mz,
            **dict_precursors_coverage
        }

        if method_conf["method_parameters"]["scan_area_A1_A2_B1_B2_only_used_for_create"][0] != int:
            next
        else:
            dict_evaluation_of_final_method["final A1, A2, B1, B2 values"] = str([
                method_conf["method_parameters"]["scan_area_A1_A2_B1_B2_only_used_for_create"][0] + method_parameters["shift_of_final_method"],
                method_conf["method_parameters"]["scan_area_A1_A2_B1_B2_only_used_for_create"][1] + method_parameters["shift_of_final_method"],
                method_conf["method_parameters"]["scan_area_A1_A2_B1_B2_only_used_for_create"][2] + method_parameters["shift_of_final_method"],
                method_conf["method_parameters"]["scan_area_A1_A2_B1_B2_only_used_for_create"][3] + method_parameters["shift_of_final_method"]]
            )
        final_df = pd.DataFrame({
            "evaluation parameter": list(dict_evaluation_of_final_method.keys()),
            "value": list(dict_evaluation_of_final_method.values())
        })
        final_df.to_csv(
            os.path.join(
                method_conf["input"]["save_at"],
                'final_method',
                'parameters_to_evaluate_method.csv'
            ),
            index=False
        )
        method_eval_table.value = final_df

        print(os.path.join(
            method_conf["input"]["save_at"],
            'final_method',
            'Kernel_density_distribution_and_final_method.png'
        ))
        self.layout[3][0] = pn.pane.PNG(
            object=os.path.join(
                method_conf["input"]["save_at"],
                'final_method',
                'Kernel_density_distribution_and_final_method.png'
            ),
            height=345,
            width=460,
            align='center',
            margin=(0, 20, 0, 0)
        )
        self.layout[3][1] = method_eval_table

def get_css_style(
    file_name="dashboard_style.css",
    directory=STYLE_PATH
):
    file = os.path.join(
        directory,
        file_name
    )
    with open(file) as f:
        return f.read()

def init_panel():
    pn.extension(raw_css=[get_css_style()])
    # pn.extension('plotly')


class GUI(object):
    # TODO: move to alphabase and docstring

    def __init__(
        self,
        name,
        img_folder_path,
        github_url,
        run_in_background=False,
        automatic_close=True,
    ):
        self.name = name
        self.tab_counter = 0
        self.header = HeaderWidget(
            name,
            img_folder_path,
            github_url
        )
        self.layout = pn.Column(
            self.header.create(),
            sizing_mode='stretch_width',
            min_width=1270
        )
        self.run_in_background = run_in_background
        self.automatic_close = automatic_close

    def start_server(self, run_in_background=False):
        if self.automatic_close:
            bokeh_ws_handler = bokeh.server.views.ws.WSHandler
            self.bokeh_server_open = bokeh_ws_handler.open
            bokeh_ws_handler.open = self.__open_browser_tab(
                self.bokeh_server_open
            )
            self.bokeh_server_on_close = bokeh_ws_handler.on_close
            bokeh_ws_handler.on_close = self.__close_browser_tab(
                self.bokeh_server_on_close
            )
        self.server = self.layout.show(threaded=True, title=self.name)
        if not run_in_background:
            self.server.join()
        elif not self.run_in_background:
            self.server.join()

    def __open_browser_tab(self, func):
        def wrapper(*args, **kwargs):
            self.tab_counter += 1
            return func(*args, **kwargs)
        return wrapper

    def __close_browser_tab(self, func):
        def wrapper(*args, **kwargs):
            self.tab_counter -= 1
            return_value = func(*args, **kwargs)
            if self.tab_counter == 0:
                self.stop_server()
            return return_value
        return wrapper

    def stop_server(self):
        logging.info("Stopping server...")
        self.server.stop()
        if self.automatic_close:
            bokeh_ws_handler = bokeh.server.views.ws.WSHandler
            bokeh_ws_handler.open = self.bokeh_server_open
            bokeh_ws_handler.on_close = self.bokeh_server_on_close


class DiAIDPasefGUI(GUI):
    # TODO: docstring

    def __init__(self, start_server=False):
        super().__init__(
            name=f"py_diAID {pydiaid.__version__}",
            img_folder_path=IMG_PATH,
            github_url='https://github.com/MannLabs/pydiaid',
        )
        self.project_description = """#### py_diAID uses an Automated Isolation Design to generate optimal dia-PASEF methods with respect to the precursor density. It designs isolation windows with variable widths, which enable short acquisition cycles, while essentially covering the complete m/z-ion mobility-range."""
        self.load_library_description = "#### Please load the library for the indicated analysis software’s to check the distribution of the precursors in the m/z-ion mobility plane."
        self.specify_parameter_description = "####  We found a strong correlation between a high theoretical and empirical precursor coverage. This result suggests using a scan area with a wide m/z-range and a narrow ion mobility range. Specify the number of dia-PASEF scans, which depend on the chromatographic peak width, and the number of ion mobility windows per dia-PASEF scan. We recommend two ion mobility windows per dia-PASEF scan."
        self.optimization_description = "#### py_diAID uses a Bayesian optimization following a Gaussian process to find the optimal scan area."
        self.create_method_description = "#### Create a dia-PASEF method with an optimal or an individually specified scan area."
        self.evaluate_method_description = "#### Evaluate the optimal dia-PASEF method or confirm if an already existing dia-PASEF method is suitable for your experiment."
        self.manual_path = os.path.join(
            DOCS_PATH,
            "manual.pdf"
        )
        self.main_widget = MainWidget(
            self.project_description,
            self.manual_path
        )

        # ERROR/WARNING MESSAGES
        self.error_message_upload = "The selected file can't be uploaded. Please check the instructions for data uploading."

        self.data = LoadLibraryCard(self.load_library_description)
        self.method_parameters = SpecifyParametersCard(
            self.data,
            self.specify_parameter_description
        )
        self.optimization = OptimizationCard(
            self.data,
            self.optimization_description
        )
        self.method_creation = CreateMethodCard(
            self.data,
            self.optimization,
            self.create_method_description
        )
        self.method_evaluation = EvaluateMethodCard(
            self.data,
            self.method_creation,
            self.evaluate_method_description
        )
        self.layout += [
            self.main_widget.create(),
            self.data.create(),
            self.method_parameters.create(),
            self.optimization.create(),
            self.method_creation.create(),
            self.method_evaluation.create(),
        ]
        if start_server:
            self.start_server()


def run():
    init_panel()
    DiAIDPasefGUI(start_server=True)


if __name__ == '__main__':
    run()



from pydiaid import method_creator
from pydiaid import method_evaluation
from pydiaid import graphs

# for data manipulation
import pandas as pd

# importing for scientific and numeric manipulations
import numpy as np
import skopt


def optimization(
    library: pd.DataFrame,
    method_parameters: dict,
    xi: np.ndarray,
    yi: np.ndarray,
    zi: np.ndarray,
    method_conf: dict,
    optimizer_parameters: dict,
) -> 'scipy.optimize.optimize.OptimizeResult':
    """ Optimize dia-PASEF method parameters: Bayesian optimization using a
    Gaussian process. The output values should follow a multivariate gaussian
    curve.

    Parameters:
    library (pd.DataFrame): a pre-filtered data frame with unified column names
        containing all required precursor information.
    method_parameters (dict): dictionary, which includes all input parameters
        for creating a dia-PASEF method.
    xi, yi, zi (numpy.ndarray): coordinates of the kernel density estimation
        where zi indicates the density
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.
    optimizer_parameters (dict): dictonary, which includes all input parameters
        for optimizing a dia-PASEF method in dependence of a proteomics library.

    Returns:
    scipy.optimize.optimize.OptimizeResult: this parameter contains important
    information regarding the optimization process (e.g., start settings,
    tested parameters and test results)
    """

    opt_result = skopt.gp_minimize(
        lambda dim: single_optimization_run(
            library,
            method_parameters,
            optimizer_parameters["evaluation_parameter"],
            dim,
            xi,
            yi,
            zi,
            method_conf
        ), [
            (optimizer_parameters["YA1"][0], optimizer_parameters["YA1"][1]),
            (
            optimizer_parameters["YA2"][0]-optimizer_parameters["YA1"][0],
            optimizer_parameters["YA2"][1]-optimizer_parameters["YA1"][0]
            ),
            (
            optimizer_parameters["YB1"][0]-optimizer_parameters["YA1"][0],
            optimizer_parameters["YB1"][1]-optimizer_parameters["YA1"][0]
            ),
            (
            optimizer_parameters["YB2"][0]-optimizer_parameters["YA1"][0],
            optimizer_parameters["YB2"][1]-optimizer_parameters["YA1"][0]
            )
        ],
        #n_random_starts = optimizer_parameters["n_random_starts"],
        n_calls = optimizer_parameters["n_calls"],  # n to test
        #optimizer_parameters["n_start"],  # random starts
        n_initial_points = optimizer_parameters["initial_points"],  # initial points
        )


    optimized_results = [opt_result.x[0],
        opt_result.x[0]+opt_result.x[1],
        opt_result.x[0]+opt_result.x[2],
        opt_result.x[0]+opt_result.x[3]]

    print("########")
    print("BEST RESULT")
    print("INPUT: " + str(
        [opt_result.x[0],
        opt_result.x[0]+opt_result.x[1],
        opt_result.x[0]+opt_result.x[2],
        opt_result.x[0]+opt_result.x[3]]
        ))
    print("OUTPUT: " + str(1.0 / opt_result.fun))
    print("########")

    write_tested_parameters_into_csv(
        opt_result,
        optimizer_parameters,
        method_conf
        )

    # return opt_result
    return optimized_results


def single_optimization_run(
    library: pd.DataFrame,
    method_parameters: dict,
    evaluation_parameter: str,
    dim: list,
    xi: np.ndarray,
    yi: np.ndarray,
    zi: np.ndarray,
    method_conf: dict,
) -> int:
    """This function contains one optimization step. First a dia-PASEF scheme
        is calculated based on the scan area coordinates, second the evaluation
        parameter is calculated to estimate the optimization potential, last
        the dia-PASEF method scheme is printed on top of the kernel density
        estimation of the used proteomics library to follow the optimization
        process visually.

    Parameters:
    library (pd.DataFrame): a pre-filtered data frame with unified column names
        containing all required precursor information.
    method_parameters (dict): dictionary, which includes all input parameters
        for creating a dia-PASEF method.
    evaluation_parameter (str): parameter, which was selected for optimization
        (e.g., all precursors, all doubly charged precursors).
    dim (list): y-coordinates of the scan area = A1, A2, B1, B2. x-coordinate
        for A1 and B1 is the lower m/z-range value, x-coordinate for A2 and B2
        is the upper m/z-range value.
    xi, yi, zi (numpy.ndarray): coordinates of the kernel density estimation
        where zi indicates the density.
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.

    Returns:
    int: the value of the parameter, that was selected for optimization (e.g.,
        all precursors, all doubly charged precursors).
    """
    # create the diaPASEF method scheme; dim[0],dim[1],dim[2],dim[3] == A1, A2,
    # B1, B2. Calculate the optimization parameter.
    library_reduced = library[library['Charge'] != 1]

    df_parameters_final = method_creator.method_creation(
        library_reduced["mz"],
        method_parameters,
        dim[0],
        dim[0]+dim[1],
        dim[0]+dim[2],
        dim[0]+dim[3]
    )
    dict_precursors_coverage = method_evaluation.coverage(
        df_parameters_final,
        library
    )
    result = 1.0 / float(dict_precursors_coverage[evaluation_parameter])
    print("RUN WITH: " + str(
    [dim[0],
    dim[0]+dim[1],
    dim[0]+dim[2],
    dim[0]+dim[3]]
    ) + " | RESULT: " + str(1.0 / result))

    # plot the created diaPASEF methods on top of the kernel density estimation
    file_name_optimization_plot = method_conf["input"]["save_at"] + '/optimization_plots/Optimization_plot_A1_' + str(dim[0]) + "_A2_ " +str(dim[0]+dim[1]) + "_B1_ " + str(dim[0]+dim[2]) + "_B2_ " + str(dim[0]+dim[3]) + "_result_" + str(1.0 / result) + ".png"
    graphs.plot_density_and_method(
        df_parameters_final,
        xi,
        yi,
        zi,
        method_conf["graphs"],
        file_name_optimization_plot
    )
    return result


def write_tested_parameters_into_csv(
    opt_result: pd.DataFrame,  # scipy.optimize.optimize.OptimizeResult
    optimizer_parameters: str,
    method_conf: dict,
) -> None:
    """This function writes all important parameters that were used during the
    method optimization step into a .csv file.

    Parameters:
    opt_result (pd.DataFrame): a dictionary containing all input and output
        information used and created by the function skopt.gp_minimize.
    optimizer_parameters (str): parameter, which was selected for optimization
        (e.g., all precursors, all doubly charged precursors).
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.

    """
    df_tested_parameters = pd.DataFrame({
        "Tested_A1_A2_B1_B2": opt_result.x_iters,
        "result": opt_result.func_vals
    })
    df_tested_parameters[optimizer_parameters["evaluation_parameter"]] = df_tested_parameters["result"].apply(lambda x: 1.0 / x)
    df_tested_parameters.to_csv(
        method_conf["input"]["save_at"] + '/optimization_plots/tested_polygon_parameters.csv',
        index=False)



# for data manipulation
import pandas as pd

# importing for scientific and numeric manipulations
import numpy as np

import os


def method_creation(
    library_mz_values: list,
    method_parameters: dict,
    A1: float,
    A2: float,
    B1: float,
    B2: float,
) -> pd.DataFrame:
    """This function calculates the dia-PASEF window coordinates based on the
        input parameters with dynamic isolation widths, adjusted to the
        precursor density in the m/z dimension. All dia-PASEF windows together
        add up to the dia-PASEF window scheme.

    Parameters:
    library_mz_values (list): list of all mz precursor values.
    method_parameters (dict): dictionary, which contains all input parameters
        for creating a dia-PASEF method (e.g., m/z-range, ion mobility-range,
        number of dia-PASEF scans, number of ion mobility steps, the overlap of
        the diaPASEF windows, the shift of the final window scheme in the ion
        mobility axis to account for the blind spot of the changing quadrupole
        at the top of each dia-PASEF window).
    A1 (float): bottom left y-coordinate of the rectangle, which defines the
        scan area and the slope of the dia-PASEF scan line.
    A2 (float): bottom right y-coordinate of the rectangle, which defines the
        scan area and the slope of the diaP-ASEF scan line.
    B1 (float): top left y-coordinate of the rectangle, which defines the scan
        area and the slope of the dia-PASEF scan line.
    B2 (float): top right y-coordinate of the rectangle, which defines the scan
        area and the slope of the dia-PASEF scan line.

    Returns:
    pd.DataFrame: data frame that contains the scan type (PASEF), scan number,
        and the corresponding diaPASEF window coordinates for each window per
        scan.
    """

    # No. of diaPASEF windows
    num_splits = method_parameters["im_steps"] * method_parameters["num_dia_pasef_scans"]

    # calculate the start and end mz range for each diaPASEF window
    # [400,420][420,450], ... :
    x_splits = divide_mz_range_in_equally_filled_parts(
        library_mz_values,
        method_parameters["mz"],
        num_splits
    )
    # calculate the position of each diaPASEF window. 1st scan: position 0, 12;
    # 2nd scan: position 1, 13:
    window_position = [[
        scan,
        method_parameters["num_dia_pasef_scans"] *
        (method_parameters["im_steps"]-1) + scan]
        for scan in range(0, method_parameters["num_dia_pasef_scans"], 1)
    ]

    m_top, c_top = line_equation(
        (method_parameters["mz"][0], B1),
        (method_parameters["mz"][1], B2)
    )
    m_bottom, c_bottom = line_equation(
        (method_parameters["mz"][0], A1),
        (method_parameters["mz"][1], A2)
    )

    results_sorted = calculate_diaPASEF_window_coordinates(
        method_parameters["im_steps"],
        method_parameters["num_dia_pasef_scans"],
        window_position,
        m_bottom,
        c_bottom,
        m_top,
        c_top,
        x_splits
    )

    return enhance_diaPASEF_windows_to_edges_of_IM_ranges(
        results_sorted,
        method_parameters["ion_mobility"],
        method_parameters["overlap"]
    )


def calculate_diaPASEF_window_coordinates(
    num_steps: int,
    num_scans: int,
    window_position: list,
    m_bottom: float,
    c_bottom: float,
    m_top: float,
    c_top: float,
    x_splits: list,
) -> pd.DataFrame:
    """This function calculates the dia-PASEF window parameters. The x start and
        end coordinate of each window (= rectangle) is already defined in
        x_splits, and the y coordinates are calculated through the intersection
        point of the x coordinates with the respective line equation.

    Parameters:
    num_steps (int): number of ion mobility windows per dia-PASEF scan.
    num_scans (int): number of diaPASEF scans.
    window_position (list(int)): position of dia-PASEF windows for each dia-PASEF
        scan. E.g., 3 dia-PASEF scans * 2 ion mobility windows per dia-PASEF scan
        = 6 dia-PASEF windows: [[0,3],[1,4],[2,5]], the first list describes the
        window positions of the first dia-PASEF scan.
    m_bottom (float): the slope of the line equation formed by A1 and A2;
        corresponds to the slope of the bottom dia-PASEF window border of the
        last ion mobility windows per dia-PASEF scan.
    c_bottom (float): y-axis intersection of the line equation formed by A1 and
        A2.
    m_top (float): the slope of the line equation formed by B1 and B2;
        corresponds to the slope of the top dia-PASEF window border of the first
        ion mobility windows per dia-PASEF scan.
    c_top (float): y-axis intersection of the line equation formed by B1 and B2.
    x_splits (list(float)): start and end m/z value of each dia-PASEF window
        sorted after dia-PASEF window position.
        E.g., [[408.2423383, 484.3129605], [484.3129605, 506.2626105],
        [506.2626105, 522.7934549], ...]

    Returns:
    pd.DataFrame: data frame containing the scan type (PASEF), scan number, and
        the corresponding dia-PASEF window coordinates for each window per scan.
    """
    y_increments = dict()
    results = list()
    scan_bottom = dict()

    for i in range(0, num_steps, 1):
        for scan in range(0, num_scans, 1):
            temp = list()
            if i == 0:
                x_bottom = x_splits[window_position[scan][0]][0]
                y_bottom = m_bottom*x_bottom + c_bottom
                x_top = x_splits[window_position[scan][1]][-1]
                y_top = m_top*x_top + c_top
                y_increments[scan] = (y_top - y_bottom) / num_steps
                scan_bottom[scan] = y_bottom

            y_bottom = scan_bottom[scan]
            y_upper = y_bottom + y_increments[scan]

            temp.append("PASEF")
            temp.append(scan + 1)
            temp.append(round(y_bottom, 2))
            temp.append(round(y_upper, 2))
            temp.append(round(x_splits[i*num_scans + scan][0], 2))
            temp.append(round(x_splits[i*num_scans + scan][-1], 2))
            scan_bottom[scan] = y_upper
            results.append(temp)

    results_sorted_decreasing_mz_value = sorted(
        results,
        key=lambda x: float(x[4]),
        reverse=True
    )
    results_sorted_increasing_diaPASEF_scan_No = sorted(
        results_sorted_decreasing_mz_value,
        key=lambda x: int(x[1])
    )
    return results_sorted_increasing_diaPASEF_scan_No


def enhance_diaPASEF_windows_to_edges_of_IM_ranges(
    df_without_column_names: pd.DataFrame,
    ion_mobility: tuple,
    overlap: float,
) -> pd.DataFrame:
    """This function recalculates the upper border of the first ion mobility
        windows per dia-PASEF scan and the lower border of the last ion mobility
        windows per dia-PASEF scan to enhance the dia-PASEF windows to the
        borders of the ion mobility range. This enhances the overall precursor
        coverage without loss in sensitivity or acquisition speed.

    Parameters:
    df_without_column_names (pd.DataFrame): data frame containing the scan type
        (PASEF), scan number, and the corresponding dia-PASEF window coordinates
        for each window per scan.
    ion_mobility (tuple): lower and upper value of the ion mobility range.
    overlap (float): this value indicates how much the dia-PASEF windows should
        overlap in the m/z dimension (unit: Da).

    Returns:
    pd.DataFrame: data frame that contains the scan type (PASEF), scan number,
        and the corresponding dia-PASEF window coordinates for each window per
        scan extended to the ion mobility borders.
    """
    df = pd.DataFrame(
        df_without_column_names,
        columns = ["MS Type", "Cycle Id", "Start IM", "End IM", "Start Mass", "End Mass"]
    )

    count = dict()
    for item in df["Cycle Id"].values:
        count[item] = count.get(item, 0)+1

    results = list()
    for i in count.keys():
        df_temp = df[df["Cycle Id"] == i]
        df_temp["End IM"].iloc[0] = ion_mobility[1]
        df_temp["Start IM"].iloc[count[i]-1] = ion_mobility[0]
        df_temp["End Mass"] = df_temp["End Mass"].apply(
            lambda x: float(x) + overlap
        )
        results.append(df_temp)
    results_concatenated = pd.concat(
        results,
        ignore_index=True
    ).values.tolist()

    return pd.DataFrame(
        results_concatenated,
        columns=["MS Type", "Cycle Id", "Start IM", "End IM", "Start Mass", "End Mass"]
    )


def divide_mz_range_in_equally_filled_parts(
    x_series: list,
    mz_range: tuple,
    num_splits: int,
) -> list:
    """This function calculates the lower and upper m/z value of each ion
        mobility window equalizing the number of precursors per diaPASEF window
        across the m/z dimension.

    Parameters:
    x_series (list(float)): list of all m/z precursor values.
    mz_range (tuple): lower and upper value of the m/z range.
    num_splits (int): number of dia-PASEF windows. Calculated by multiplying the
        number of dia-PASEF scans with the number of ion mobility windows per
        diaPASEF scan.

    Returns:
    list(float): start and end m/z value of each dia-PASEF window sorted after
        diaPASEF window position. E.g., [[408.2423383, 484.3129605],
        [484.3129605, 506.2626105], [506.2626105, 522.7934549], ...]
    """
    # filter for precursors within the m/z range
    x_filtered = [ele for ele in sorted(list(x_series.values)) if mz_range[0] <= ele <= mz_range[1]]
    # reminder of division, No. of precursors, which are hard to distributed
    # equally, [0,1,2,3,4,5,6] num_split = 3 -> reminder: 1
    rem = len(x_filtered) % num_splits
    # to distribute precursors equally; the last m/z value is added to the
        # list until the division is without reminder.-> [0, 1, 2, 3, 4, 5, 6, 6, 6]
    x_filtered += [x_filtered[-1] for i in range(num_splits - rem)]

    results = list()
    for arr in np.split(np.array(x_filtered), num_splits):
    # split the list [0, 1, 2, 3, 4, 5, 6, 6, 6] in num_split parts (here: 3):
        # [0 1 2], [3 4 5], [6 6 6]; save upper and lower m/z border in results
        # -> [0,2],[3,5],[6,6]
        results.append(
            [
                arr[0],
                arr[-1]] if len(results) == 0 else [results[-1][1], arr[-1]
            ]
        )
    return results


def line_equation(
    l1: tuple,
    l2: tuple,
) -> float:
    """This function calculates the line equation of two coordinates.
    Parameters:
    l1 (tuple): x and y coordinate of the 1st point. Line encoded as l=(x,y).
    l2 (tuple): x and y coordinate of the 2nd point.

    Returns:
    m (float): the slope of the line equation.
    c (float): y-axis intersection of the line equation formed by l1 and l2.
    """
    m = float((l2[1] - l1[1])) / float(l2[0] - l1[0])
    c = (l2[1] - (m * l2[0]))
    return m, c


def create_parameter_dataframe(
    df_parameters_final: pd.DataFrame,
    file_name: str,
) -> None:
    """This function writes the dia-PASEF parameter file as .txt. This file
        serves as input for the generation of new dia-PASEF methods with
        timsControl (timsTof Pro 2 control software).
    Parameters:
    df_parameters_final (pd.DataFrame): data frame that contains the scan type
        (PASEF), scan number, and the corresponding dia-PASEF window coordinates
        for each window per scan extended to the ion mobility borders.
    file_name (str): file path and file name where the dia-PASEF parameter file
        should be saved.
    """
    results_input = df_parameters_final.astype(str)
    results = results_input.values.tolist()

    with open(os.path.join(os.path.dirname(__file__), "static/DIAParameterspy3TC.txt")) as f:
        content = f.readlines()

    ref_line = [
        len(col.replace("\n",""))
        for col in content[5].split(',')
    ][:-1]
    content_new = list()
    for result in results:
         content_new.append(
            ",".join([result[i].rjust(ref_line[i])
            for i in range(len(result))] + [content[5].split(',')[-1]])
        )
    final_content = content[:4] + content_new

    with open(file_name, 'w') as filehandle:
        filehandle.writelines(final_content)



# for data manipulation
import pandas as pd
import os
import re


def load_library(
    library_name: str,
    analysis_software: str,
    ptm: str,
) -> pd.DataFrame:
    """Loads an output file of a proteomics analysis software as a data frame,
        which should represent the diversity of possible peptides.

    Parameters:
    library_name (str): path, where the proteomics library is stored.
    analysis_software (str): an identifier for the analysis software used to
        create the input data. The script chooses different parse functions
        depending on this identifier.
    ptm (str): an identifier used for filtering a specific data frame column
        for this string.

    Returns:
    pd.DataFrame: returns a pre-filtered data frame with unified column names
        independent of the input data frame analyzed with individual analysis
        software.
    """

    try:
        dataframe = __load_dataframe_from_file(library_name)
        if analysis_software == 'AlphaPept':
            return __parse_alpha_pept(dataframe, ptm)
        if analysis_software == 'MaxQuant':
            return __parse_max_quant(dataframe, ptm)
        if analysis_software == 'FragPipe':
            return __parse_ms_fragger(dataframe, ptm)
        if analysis_software == 'Spectronaut single-run':
            return __parse_spectronaut_single_shot(dataframe, ptm)
        if analysis_software == 'Spectronaut library':
            return __parse_spectronaut_library(dataframe, ptm)
        raise Exception('Analysis software not supported.')
    except Exception as e:
        print(e)
        raise Exception("error while processing: Did you choose the correct analysis_software and is the modification present in the dataset?")


def __load_dataframe_from_file(
    library_name: str,
) -> pd.DataFrame:
    """Imports an output file of a proteomics analysis software as a data frame
        independent of the file format.

    Parameters:
    library_name (str): path, where the proteomics library is stored.

    Returns:
    pd.DataFrame: returns a data frame in the same way as the specific
        proteomics analysis software stored it.
    """

    if library_name.split(".")[-1] == "csv":
        return pd.read_csv(library_name, sep=',')
    else:
        return pd.read_csv(library_name, sep='\t')  # .xls, .tsv, .txt


def __parse_alpha_pept(
    dataframe: pd.DataFrame,
    ptm: str,
) -> pd.DataFrame:
    """Filters a data frame depending on the software specific requirements to
        only include valid precursors. Additionally, it parses the data frame
        to library_loader to filter for specific modified peptides and to unify
        the column names of the required columns.

    Parameters:
    dataframe (pd.DataFrame): imported output file from the analysis software
        "AlphaPept". File format: .csv, required columns: "q_value", "decoy",
        'mz', 'mobility', 'charge', 'protein', 'precursor'.
    ptm (str): an identifier used for filtering a specific data frame column.

    Returns:
    pd.DataFrame: returns a pre-filtered data frame with unified column names.
    """

    filtered_dataframe = dataframe[
        (dataframe["q_value"] <= 0.01) &
        (dataframe["decoy"] == False)
    ]
    library_subset = library_loader(
        filtered_dataframe,
        ptm,
        'mz',
        'mobility',
        'charge',
        'protein',
        'precursor'
    )
    return library_subset


def __parse_max_quant(
    dataframe: pd.DataFrame,
    ptm: str,
) -> pd.DataFrame:
    """Filters a data frame depending on the software specific requirements to
        only include valid precursors. Additionally, it parses the data frame
        to library_loader to filter for specific modified peptides and to unify
        the column names of the required columns.

    Parameters:
    dataframe (pd.DataFrame): imported output file from the analysis software
        "MaxQuant".
    File format: evidence.txt, required columns: "Reverse", "Potential
        contaminant", 'm/z', '1/K0', 'Charge', 'Proteins', 'Modified sequence'.
    ptm (str): an identifier used for filtering a specific data frame column.

    Returns:
    pd.DataFrame: returns a pre-filtered data frame with unified column names.
    """

    filtered_dataframe = dataframe[
        (dataframe["Reverse"] != "+") &
        (dataframe["Potential contaminant"] != "+")
    ]
    library_subset = library_loader(
        filtered_dataframe,
        ptm,
        'm/z',
        '1/K0',
        'Charge',
        'Proteins',
        'Modified sequence'
    )
    return library_subset


def __parse_ms_fragger(
    dataframe: pd.DataFrame,
    ptm: str,
) -> pd.DataFrame:
    """It parses the data frame to library_loader to filter for specific
        modified peptides and to unify the column names of the required
        columns.

    Parameters:
    dataframe (pd.DataFrame): imported output file from the analysis software
        "MSFragger".
    File format: .tsv, required columns: 'PrecursorMz', 'PrecursorIonMobility',
        'PrecursorCharge', 'ProteinId', 'ModifiedPeptideSequence'.
    ptm (str): an identifier used for filtering a specific data frame column.

    Returns:
    pd.DataFrame: returns a pre-filtered data frame with unified column names.
    """

    library_subset = library_loader(
        dataframe,
        ptm,
        'PrecursorMz',
        'PrecursorIonMobility',
        'PrecursorCharge',
        'ProteinId',
        'ModifiedPeptideSequence'
    )
    return library_subset


def combine_columns(
    x: pd.DataFrame,
) -> str:
    """One line of a data frame is parsed to this fuction to combine the
        columns 'Modified Peptide' & 'Peptide'. Only necesary in rare cases.

    Parameters:
    x (pd.DataFrame): parsed line of a dataframe to combine two columns to one.

    Returns:
    str: output value with combined information.
    """
    if pd.isna(x['Modified Peptide']) is True:
        return x['Peptide']
    else:
        return x['Modified Peptide']


def __parse_spectronaut_single_shot(
    dataframe: pd.DataFrame,
    ptm: str,
) -> pd.DataFrame:
    """Filters a data frame depending on the software specific requirements to
        only include valid precursors. Additionally, it parses the data frame
        to library_loader to filter for specific modified peptides and to unify
        the column names of the required columns.

    Parameters:
    dataframe (pd.DataFrame): imported output file from the analysis software
        "Spectronaut - DIA analysis".
    File format: .xls, required columns: 'PG.ProteinGroups', 'EG.PrecursorId',
        'FG.PrecMzCalibrated', 'FG.ApexIonMobility', 'FG.Charge'.
    ptm (str): an identifier used for filtering a specific data frame column.

    Returns:
    pd.DataFrame: returns a pre-filtered data frame with unified column names.
    """

    filter_selection = (dataframe['FG.PrecMzCalibrated'].isnull() == False)
    filter_selection &= (dataframe['FG.ApexIonMobility'].isnull() == False)
    filtered_library = dataframe[filter_selection]
    library_subset = library_loader(
        filtered_library,
        ptm,
        'FG.PrecMzCalibrated',
        'FG.ApexIonMobility',
        'FG.Charge',
        'PG.ProteinGroups',
        'EG.PrecursorId'
    )
    return library_subset


def __parse_spectronaut_library(
    dataframe: pd.DataFrame,
    ptm: str,
) -> pd.DataFrame:
    """Filters a data frame depending on the software specific requirements to
        only include valid precursors. Additionally, it parses the data frame
        to library_loader to filter for specific modified peptides and to unify
        the column names of the required columns.

    Parameters:
    dataframe (pd.DataFrame): imported output file from the analysis software
        "Spectronaut - library generation from Pulsar".
    File format: .xls, required columns: 'PrecursorMz', 'IonMobility',
        'PrecursorCharge', 'UniProtIds', 'ModifiedPeptide'.
    ptm (str): an identifier used for filtering a specific data frame column.

    Returns:
    pd.DataFrame: returns a pre-filtered data frame with unified column names.
    """
    filter_selection = (dataframe['PrecursorMz'].isnull() == False)
    filter_selection &= (dataframe['IonMobility'].isnull() == False)
    filtered_library = dataframe[filter_selection]

    library_subset = library_loader(
        filtered_library,
        ptm,
        'PrecursorMz',
        'IonMobility',
        'PrecursorCharge',
        'UniProtIds',
        'ModifiedPeptide'
    )

    return library_subset


def library_loader(
    library: pd.DataFrame,
    ptm: str,
    mz: str,
    im: str,
    charge: str,
    protein: str,
    modified_peptide: str,
) -> pd.DataFrame:
    """Filters a column of a data frame for a specific identifier (e.g.,
        "Phospho") and unifies the required column names.

    Parameters:
    library (pd.DataFrame): data frame, which should be modified. It needs to
        have columns, which indicate the precursor m/z values, precursor ion
        mobility, precursor charge, corresponding proteins, and modified
        peptide sequence (can contain also charge state information)
    ptm (str): an identifier used for filtering the data frame column "modified
        peptide sequence" for a specific modification.
    mz (str): column name of the column containing the precursor m/z value.
    im (str): column name of the column containing the precursor specific ion
        mobility.
    charge (str): column name of the column containing the precursor charge
        state information.
    protein (str): column name of the column stating the corresponding proteins
        to a specific precursor.
    modified_peptide (str):  column name containing the modified peptide
        sequence for each precursor (can additionally contain charge state
        information).

    Returns:
    pd.DataFrame: returns a pre-filtered data frame with unified column names.
    """
    if ptm != 'None':
        library[ptm] = library[modified_peptide].apply(
            lambda x: find_PTM(x, ptm)
        )
        try:
            library_filtered = library[library[ptm] == True]
            raise Exception('Modification not in dataset.')
        except Exception as e:
            print(e)
            raise Exception("Is this modification present in the dataset?")
        library_subset = library_filtered.drop_duplicates(
            [modified_peptide, charge]
        )
    if ptm == 'None':
        library_subset = library.drop_duplicates(
            [modified_peptide, charge]
        )  # use only unique precursors

    library_small = pd.DataFrame()
    library_small['mz'] = library_subset[mz]
    library_small['IM'] = library_subset[im]
    library_small['Charge'] = library_subset[charge]
    library_small['Proteins'] = library_subset[protein]
    library_small['Peptide'] = library_subset[modified_peptide]
    return library_small


def find_PTM(column_value, ptm):
    """Identifies if a column value indicates a modification.

    Parameters:
    column_value (str): the value of the column.
    PTM (str): name of the item which is searched for, possibilities:
        'Phospho', 'STY' or 'Gly'

    Returns:
    boolean: Indicates if the specific string (ptm) is present in the column
        value, which allows filtering for modified peptides.
    """

    if ptm in column_value:
        return True
    else:
        next


def get_file_names_from_directory(
    directory: str,
    extensions_list: list
)-> list:
    """Search for files with the specified extension in the repository and
    return a list of all file names with that extention.

    Parameters
    ----------
    directory : str
        Path to the repository to search in.
    extensions_list : list
        A list of extensions, e.g. ['d', 'hdf'].

    Returns
    -------
    list
        The list of filtered file names based on their extensions.
    """
    file_names = [file for file in os.listdir(directory) if file.split('.')[-1] in extensions_list]
    return file_names


def create_opt_plot_df(
    filename: str
)-> pd.DataFrame:
    """Return the dataframe containing information about the optimized scan
    area coordinates and coverage based on the filename of the .png file.

    Parameters
    ----------
    filename : str
        The name of the .png file showing the kernel density estimation.

    Returns
    -------
    pd.DataFrame
        The data frame contains several columns:
            - parameters: ['A1', 'A2', 'B1', 'B2', 'coverage'];
            - values: showing the values for each parameter extracted from the
            filename.

    """

    values = re.findall(r'(\d+\.\d+)', filename)
    df = pd.DataFrame(
        {
            'parameters': ['A1', 'A2', 'B1', 'B2', 'coverage'],
            'values': values
        }
    )
    return df



# for data manipulation
import pandas as pd

# importing for scientific and numeric manipulations
import numpy as np

# importing components for visualization
from matplotlib.patches import Rectangle


def calculate_precursor_within_scan_area(
    library: pd.DataFrame,
    mz: tuple,
    im: tuple,
) -> dict:
    """This function calculates the number of precursors within the scan area
    in percent.

    Parameters:
    library (pd.DataFrame): a pre-filtered data frame with unified column names
        containing all required precursor information.
    mz (tuple): lower and upper value of the m/z range.
    im (tuple): lower and upper value of the ion mobility range.

    Returns:
    dict: this dictionary has only one item. The key is the value description,
        and the value is the ratio of precursors [%] within the set scan area.
    """
    filtered_selection = library['mz'] > mz[0]
    filtered_selection &= library['mz'] < mz[1]
    filtered_selection &= library['IM'] > im[0]
    filtered_selection &= library['IM'] < im[1]

    prec_scan_area = (len(library[filtered_selection])/len(library['mz']))*100

    return {'precursors within m/z-range [%]': round(prec_scan_area, 2)}


def calculate_percentage_multiple_charged_precursors(
    library_subset: pd.DataFrame,
) -> dict:
    """This function calculates the ratio of precursors with different charge
        states in percent.

    Parameters:
    library_subset (pd.DataFrame): a pre-filtered data frame with unified
        column names containing all required precursor information.

    Returns:
    dict: The key is the value description, and the value is the ratio of
        precursors [%] with a specific charge state.
    """
    dict_charge_prec = dict()
    for i in range(1, 5):
        dict_charge_prec[str(i)] = round(
            (
                len(
                    library_subset[library_subset['Charge'] == i]
                ) / len(library_subset)
            ) * 100, 2
        )
    return dict_charge_prec


def coverage(
    df: pd.DataFrame,
    library: pd.DataFrame,
) -> dict:
    """This function calculates different evaluation values to estimate the
        optimization potential of the tested dia-PASEF method based on the
        theoretically covered (doubly/ triply charged) precursors.

    Parameters:
    df (pd.DataFrame): data frame that contains the scan type (PASEF), scan
        number, and the corresponding dia-PASEF window coordinates for each
        window per scan.
    library_subset (pd.DataFrame): a pre-filtered data frame with unified
        column names containing all required precursor information.

    Returns:
    dict: This dictionary describes multiple values to evaluate diaPASEF
        methods in respect to the proteomics library.
    """
    _, borders, small_window, big_window, mean_window = boxes(df)

    coverage_list = list()
    for i in borders:
        # subset for precursors within boundaries and append to dataset:
        # mz_min:i[0], mz_max:i[2], im_min:i[1], im_max:i[3]
        filtered_selection = library['mz'] >= i[0]
        filtered_selection &= library['mz'] <= i[2]
        filtered_selection &= library['IM'] >= i[1]
        filtered_selection &= library['IM'] <= i[3]
        coverage_list.append(library[filtered_selection])
    coverage_unique_prec = pd.concat(coverage_list, ignore_index=True)
    coverage_removed_prec_duplicates = coverage_unique_prec.drop_duplicates(['Peptide', 'Charge'])

    dict_prec_coverage = {
        "unique proteins in the library": len(library.drop_duplicates(['Proteins'])),
        "unique precursors in the library": len(library.drop_duplicates(['Peptide', 'Charge'])),
        "smallest diaPASEF window": round(small_window, 2),
        "biggest diaPASEF window": round(big_window, 2),
        "average diaPASEF window size": round(mean_window, 2),
        "No. of covered proteins": len(coverage_removed_prec_duplicates.drop_duplicates(['Proteins'])),
        "No. of covered precursors": len(coverage_removed_prec_duplicates),
        "No. of covered, doubly charged precursors": len(coverage_removed_prec_duplicates[coverage_removed_prec_duplicates['Charge'] == 2]),
        "No. of covered, triply charged precursors": len(coverage_removed_prec_duplicates[coverage_removed_prec_duplicates['Charge'] == 3]),
        "No. of covered, quadruply charged precursors": len(coverage_removed_prec_duplicates[coverage_removed_prec_duplicates['Charge'] == 4])
    }
    dict_prec_coverage["all proteins covered"] = format(np.round(dict_prec_coverage["No. of covered proteins"]/len(library.drop_duplicates(['Proteins']))*100, 1)) + "%"
    dict_prec_coverage["all precursors covered"] = format(np.round(dict_prec_coverage["No. of covered precursors"]/len(library)*100, 1)) + "%"
    dict_prec_coverage["all doubly charged precursors covered"] = format(np.round(dict_prec_coverage["No. of covered, doubly charged precursors"] / len(library[library['Charge'] == 2]) * 100, 1)) + "%"
    dict_prec_coverage["all triply charged precursors covered"] = format(np.round(dict_prec_coverage["No. of covered, triply charged precursors"] / len(library[library['Charge'] == 3]) * 100, 1)) + "%"
    dict_prec_coverage["all quadruply charged precursors covered"] = format(np.round(dict_prec_coverage["No. of covered, quadruply charged precursors"] / len(library[library['Charge'] == 4]) * 100, 1)) + "%"
    return dict_prec_coverage


def boxes(
    df: pd.DataFrame,
) -> int:
    """This function calculates dia-PASEF window specific information regarding
        the window size and creates two lists. One is for plotting the dia-PASEF
        window scheme on top of a kernel density estimation plot. The second is
        to calculate the number of proteins and precursors covered by the
        dia-PASEF window scheme.

    Parameters:
    df (pd.DataFrame): data frame that contains the scan type (PASEF), scan
        number, and the corresponding dia-PASEF window coordinates for each
        window per scan.

    Returns:
    rect (list): This list contains all information to print the dia-PASEF
        windows on top of a kernel density estimation plot.
    border (list(tuples)): The list "borders" contains a list of tuples where
        each tuple represents the edge coordinates for each dia-PASEF window.
    small_window (int): size of the smallest dia-PASEF window [Da].
    big_window (int): size of the biggest dia-PASEF window [Da].
    mean_window (int): average dia-PASEF window size [Da].
    """
    rect = list()
    borders = list()
    window_width = list()
    count = dict()

    for item in df["Cycle Id"].values:
        count[item] = count.get(item, 0)+1

    for i in count.keys():
        df_temp = df[df["Cycle Id"] == i]
        df_temp["xw"] = df_temp["End Mass"] - df_temp["Start Mass"]
        df_temp["yw"] = df_temp["End IM"] - df_temp["Start IM"]

        df_temp["borders"] = df_temp.apply(
            lambda x:
                (x["Start Mass"], x["Start IM"], x["End Mass"], x["End IM"]),
                axis=1
            )
        df_temp["rect"] = df_temp.apply(
            lambda x: Rectangle(
                (x["Start Mass"], x["Start IM"]), x["xw"], x["yw"]),
                axis=1
            )
        rect += list(df_temp["rect"].values)
        borders += list(df_temp["borders"].values)
        window_width += list(df_temp["xw"].values)
        small_window = min(window_width)
        big_window = max(window_width)
        mean_window = np.mean(window_width)
    return rect, borders, small_window, big_window, mean_window



#!python

# external
import click

# local
import pydiaid
from pydiaid import loader
from pydiaid import main
from pydiaid import method_optimizer

# for data manipulation
import pandas as pd

# for saving input data as json:
import json

# for creating new folders:
import os

BASE_PATH = os.path.dirname(__file__)
LIB_PATH = os.path.join(BASE_PATH, "lib")
DEFAULT_FILE = os.path.join(
    LIB_PATH,
    "default_parameters.json"
)


@click.group(
    context_settings=dict(
        help_option_names=['-h', '--help'],
    ),
    invoke_without_command=True
)
@click.pass_context
@click.version_option(pydiaid.__version__, "-v", "--version")
def run(ctx, **kwargs):
    name = f"py_diAID {pydiaid.__version__}"
    click.echo("*" * (len(name) + 4))
    click.echo(f"* {name} *")
    click.echo("*" * (len(name) + 4))
    if ctx.invoked_subcommand is None:
        click.echo(run.get_help(ctx))


@run.command("gui", help="Start graphical user interface.")
def gui():
    import pydiaid.gui
    pydiaid.gui.run()


@run.command("optimize", help="Optimize a dia-PASEF method.")
@click.option(
    "parameter_file_name",
    "-p",
    help=f"Parameter file (check out {DEFAULT_FILE} for an example)",
    default=DEFAULT_FILE,
    required=True,
)
def optimize_window_scheme(parameter_file_name):
    print(f"Using parameter file {parameter_file_name}")
    with open(parameter_file_name, "r") as infile:
        method_conf = json.load(infile)
    run_all(method_conf)


@run.command("create", help="Create a specific dia-PASEF method.")
@click.option(
    "parameter_file_name",
    "-p",
    help=f"Parameter file (check out {DEFAULT_FILE} for an example)",
    default=DEFAULT_FILE,
    required=True,
)
def create_diaPASEF_method(parameter_file_name,):
    print(f"Using parameter file {parameter_file_name}")
    with open(parameter_file_name, "r") as infile:
        method_conf = json.load(infile)
    library_plus_create_methods(method_conf)


@run.command("evaluate", help="Evaluate a dia-PASEF method.")
@click.option(
    "parameter_file_name",
    "-p",
    help=f"Parameter file (check out {DEFAULT_FILE} for an example)",
    default=DEFAULT_FILE,
    required=True,
)
def evaluate_diaPASEF_method(parameter_file_name,):
    print(f"Using parameter file {parameter_file_name}")
    with open(parameter_file_name, "r") as infile:
        method_conf = json.load(infile)
    library_plus_evaluate_method(method_conf)


@run.command("charge", help="Evaluate a dia-PASEF method for multiply charged precursors.")
@click.option(
    "parameter_file_name",
    "-p",
    help=f"Parameter file (check out {DEFAULT_FILE} for an example)",
    default=DEFAULT_FILE,
    required=True,
)
def evaluate_multiple_charged_prec(parameter_file_name,):
    print(f"Using parameter file {parameter_file_name}")
    with open(parameter_file_name, "r") as infile:
        method_conf = json.load(infile)
    multiple_charged_prec(method_conf)


def run_all(
    method_conf: dict,
) -> None:
    """This function carries out all sub-functions step by step: creating a
    folder for the output information, loading of the proteomics library,
    calculation of the kernel density estimation for the density plots,
    Bayesian optimization of dia-PASEF method parameters by trying multiple scan
    area coordinates using a Gaussian process, writing all input and output
    information in json or csv files, creating final dia-PASEF method, plotting
    dia-PASEF windows on top of a kernel density estimation of precursors and
    calculating method evaluation information.

    Parameters:
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.

    """

    method_parameters = method_conf["method_parameters"]
    optimizer_parameters = method_conf["optimizer"]
    save_at = method_conf["input"]["save_at"]

    folder_paths = [
        save_at,
        save_at+'/optimization_plots',
        save_at+'/input_library',
        save_at+'/final_method'
        ]
    main.create_folder(folder_paths)

    # save input parameters as json file for reusing
    out_file = open(save_at+"/input_parameters.json", "w")
    json.dump(method_conf, out_file, indent=4)
    out_file.close()

    xi, yi, zi, library = main.library_information(
        method_conf,
        save_at
        )

    main.precursor_within_scan_area(
        library,
        method_parameters,
        save_at
        )

    dim = method_optimizer.optimization(
        library,
        method_parameters,
        xi,
        yi,
        zi,
        method_conf,
        optimizer_parameters
        )

    df_parameters_final = main.create_final_method(
        library,
        method_parameters,
        dim,
        method_conf
        )

    main.final_method_information(
        df_parameters_final,
        xi,
        yi,
        zi,
        method_conf,
        save_at, library,
        method_parameters,
        dim
        )


def library_plus_evaluate_method(
    method_conf: dict
) -> None:
    """This function carries out all sub-functions required for method
    evaluation: creating a folder for the output information, loading of the
    proteomics library, loading of the acquisition scheme, calculation of the
    kernel density estimation for the density plots, writing all input and
    output information in json or csv files, plotting of the diaPASEF windows
    on top of a kernel density estimation of precursors, and calculating method
    evaluation information.

    Parameters:
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.

    """

    method_parameters = method_conf["method_parameters"]
    save_at = method_conf["input"]["save_at"]

    folder_paths = [
        save_at,
        save_at+'/input_library',
        save_at+'/final_method'
        ]
    main.create_folder(folder_paths)

    # save input parameters as json file for reusing
    out_file = open(save_at+"/input_parameters.json", "w")
    json.dump(method_conf, out_file, indent=4)
    out_file.close()

    df_parameters_final = pd.read_csv(
        method_conf["input"]["diaPASEF_method_only_used_for_evaluate"],
        skiprows=4,
        names=["MS Type", "Cycle Id", "Start IM", "End IM", "Start Mass", "End Mass", "CE"]
        )

    xi, yi, zi, library = main.library_information(
        method_conf,
        save_at
        )

    main.final_method_information(
        df_parameters_final,
        xi,
        yi,
        zi,
        method_conf,
        save_at,
        library,
        method_parameters,
        None
        )

    print("Method evaluated")


def library_plus_create_methods(
    method_conf: dict
) -> None:
    """This function carries out all sub-functions required to generate a
    dia-PASEF method with specific scan area coordinates (A1, A2, B1, B2):
    creating a folder for the output information, loading of the proteomics
    library, writing all input and output information in json or csv files,
    creating the dia-PASEF method.

    Parameters:
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.

    """

    method_parameters = method_conf["method_parameters"]
    save_at = method_conf["input"]["save_at"]
    dim = method_parameters["scan_area_A1_A2_B1_B2_only_used_for_create"]

    folder_paths = [
        save_at,
        save_at+'/final_method'
        ]
    main.create_folder(folder_paths)

    # save input parameters as json file for reusing
    out_file = open(save_at+"/input_parameters.json", "w")
    json.dump(method_conf, out_file, indent=4)
    out_file.close()

    library = loader.load_library(
        method_conf["input"]["library_name"],
        method_conf["input"]["analysis_software"],
        method_conf["input"]["PTM"]
        )

    main.precursor_within_scan_area(
        library,
        method_parameters,
        save_at
        )

    main.create_final_method(
        library,
        method_parameters,
        dim,
        method_conf
        )

    print("Method created")


def multiple_charged_prec(
    method_conf: dict
) -> None:
    """This function carries out all sub-functions required for method
    evaluation of precursor with multiple charge states: creating a folder for
    the output information, loading of the proteomics library, loading of the
    acquisition scheme, calculation of the kernel density estimation for the
    density plots depending on the charge state, writing all input information
    in a json file, plotting of the diaPASEF windows on top of a kernel density
    estimation and calculating of method evaluation information.

    Parameters:
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.

    """

    save_at = method_conf["input"]["save_at"]

    folder_paths = [
        save_at,
        save_at+'/evaluation_multiple_charged_prec'
        ]
    main.create_folder(folder_paths)

    # save input parameters as json file for reusing
    out_file = open(save_at+"/input_parameters.json", "w")
    json.dump(method_conf, out_file, indent=4)
    out_file.close()

    df_parameters_final = pd.read_csv(
        method_conf["input"]["diaPASEF_method_only_used_for_evaluate"],
        skiprows=4,
        names=["MS Type", "Cycle Id", "Start IM", "End IM", "Start Mass", "End Mass", "CE"]
        )

    library = loader.load_library(
        method_conf["input"]["library_name"],
        method_conf["input"]["analysis_software"],
        method_conf["input"]["PTM"]
        )

    main.evaluate_for_multiple_charged_prec(
        method_conf,
        save_at,
        library,
        df_parameters_final
    )



# for data manipulation
import pandas as pd

# importing for scientific and numeric manipulations
from scipy.stats import kde
import numpy as np

# importing components for visualization
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.collections import PatchCollection

# for suppressing warnings
import warnings

from pydiaid.method_evaluation import boxes

# importing components for visualization
mpl.rcParams['font.family'] = 'Arial'
plt.rcParams.update({'font.size': 16.5})
mpl.rcParams['pdf.fonttype'] = 42

# for suppressing warnings
warnings.filterwarnings('ignore')


def kernel_density_calculation(
    library_subset: pd.DataFrame,
    nbins: int,
) -> np.ndarray:
    """Calculate the kernel density estimation of a data frame representing a
        filtered proteomics library or single-shot measurement.

    Parameters:
    library_subset (pd.DataFrame): pre-filtered data frame with unified column
        names.
    nbins (int): number of bins for the kernel density estimation.

    Returns:
    xi, yi, zi (numpy.ndarray): coordinates of the kernel density estimation,
        where zi indicates the density.
    """
    x = library_subset['mz']
    y = library_subset['IM']

    k = kde.gaussian_kde([x, y])
    xi, yi = np.mgrid[x.min():x.max():nbins*1j, y.min():y.max():nbins*1j]
    zi = k(np.vstack([xi.flatten(), yi.flatten()]))

    return xi, yi, zi


def plot_density(
    xi: np.ndarray,
    yi: np.ndarray,
    zi: np.ndarray,
    plot_parameters: dict,
    file_name: str,
    gui: bool = False
) -> None:
    """Create a density plot from a data frame representing a filtered
        proteomics library or single-shot measurement.

    Parameters:
    xi, yi, zi (numpy.ndarray): coordinates of the kernel density estimation,
        where zi indicates the density.
    plot_parameters (dict): dictionary, which contains all input parameters for
        creating plots (e.g., displayed m/z-range, ion mobility-range).
    file_name: file path and file name where the plot should be saved
    gui (bool): whether to use in the GUI or not. Defaults is False.
    """
    fig, ax = plt.subplots()
    plt.pcolormesh(
        xi, yi, zi.reshape(xi.shape),
        vmax=0.02,
        cmap=plt.cm.viridis
    )
    ax.set_xlim(plot_parameters["plot_mz"][0], plot_parameters["plot_mz"][1])
    ax.set_ylim(plot_parameters["plot_IM"][0], plot_parameters["plot_IM"][1])
    ax.set_facecolor('#440256')
    plt.xlabel('$\mathregular{\it{m/z}}$')
    plt.ylabel('$\mathregular{1/K_0}$ [Vs $\mathregular{cm^{-2}}$]')
    plt.colorbar().set_label('Density', labelpad=-28, y=1.14, rotation=0)
    plt.savefig(file_name, bbox_inches='tight', pad_inches=0, dpi=300)
    if gui:
        return fig
    else:
        plt.clf()


def plot_precursor_distribution_as_histogram(
    library_subset: pd.DataFrame,
    plot_parameters: dict,
    file_name: str,
    gui: bool = False
) -> None:
    """Plot histogram with the precursor distribution in the m/z dimension
        sorted by charge state.

    Parameters:
    library_subset (pd.DataFrame): pre-filtered data frame with unified column
        names.
    plot_parameters (dict): dictionary, which contains all input parameters for
        creating plots (e.g., displayed m/z-range, ion mobility-range).
    file_name: file path and file name where the plot should be saved
    gui (bool): whether to use in the GUI or not. Defaults is False.
    """
    fig = plt.figure()

    plt.hist(
        [
            library_subset['mz'][library_subset['Charge'] == 3],
            library_subset['mz'][library_subset['Charge'] == 2],
            library_subset['mz']
        ],
        bins=100,
        histtype='step',
        fill=True,
        alpha=0.5,
        color=['#267FA5', '#4EA7BB', '#7AC7C9'],
        label=[
            'triply charged precursors',
            'doubly charged precursors',
            'all precursors'
        ]
    )

    plt.xlim([plot_parameters["plot_mz"][0], plot_parameters["plot_mz"][1]])
    plt.ylabel('No. of precursors')
    plt.legend(bbox_to_anchor=(0.8, 1.42))
    plt.xlabel('$m/z$')
    fig.savefig(file_name, bbox_inches='tight', pad_inches=0, dpi=300)
    if gui:
        plt.legend(
            loc='upper right',
            fontsize='x-small'
        )
        return fig
    else:
        plt.clf()


def plot_density_and_method(
    df: pd.DataFrame,
    xi: np.ndarray,
    yi: np.ndarray,
    zi: np.ndarray,
    plot_parameters: dict,
    file_name: str,
) -> None:
    """Plot the dia-PASEF acquisition scheme on top of a density plot
    representing a filtered proteomics library or single-shot measurement.

    Parameters:
    df (pd.DataFrame): data frame that contains the scan type (PASEF), scan
        number and the corresponding di-aPASEF window coordinates for each
        window per scan.
    xi, yi, zi (numpy.ndarray): coordinates of the kernel density estimation,
        where zi indicates the density.
    plot_parameters (dict): dictionary, which contains all input parameters for
        creating plots (e.g., displayed m/z-range, ion mobility-range).
    file_name: file path and file name where the plot should be saved.
    """

    rect, _, _, _, _ = boxes(df)

    fig, ax = plt.subplots()
    plt.pcolormesh(
        xi, yi, zi.reshape(xi.shape),
        vmax=0.02,
        cmap=plt.cm.viridis
    )
    pc = PatchCollection(
        rect,
        facecolor=plot_parameters["window_color"],
        alpha=plot_parameters["window_transparency"],
        edgecolor=plot_parameters["window_frame_color"]
    )
    ax.add_collection(pc)

    ax.set_xlim(plot_parameters["plot_mz"][0], plot_parameters["plot_mz"][1])
    ax.set_ylim(plot_parameters["plot_IM"][0], plot_parameters["plot_IM"][1])
    ax.set_facecolor('#440256')
    plt.xlabel('$\mathregular{\it{m/z}}$')
    plt.ylabel('$\mathregular{1/K_0}$ [Vs $\mathregular{cm^{-2}}$]')
    plt.colorbar().set_label('Density', labelpad=-28, y=1.14, rotation=0)
    plt.savefig(file_name, bbox_inches='tight', pad_inches=0, dpi=300)
    plt.clf()


def kernel_density_calculation_multiple_charge(
    library_subset: pd.DataFrame,
    nbins: int,
    charge_setting: int,
) -> np.ndarray:
    """Calculate the kernel density estimation for specific multiple charged
        precursors.

    Parameters:
    library_subset (pd.DataFrame): pre-filtered data frame with unified column
        names.
    nbins (int): number of bins for the kernel density estimation.

    Returns:
    xi, yi, zi (numpy.ndarray): coordinates of the kernel density estimation,
        where zi indicates the density.
    """
    x = library_subset['mz'][library_subset['Charge'] == charge_setting]
    y = library_subset['IM'][library_subset['Charge'] == charge_setting]

    k = kde.gaussian_kde([x, y])
    xi, yi = np.mgrid[x.min():x.max():nbins*1j, y.min():y.max():nbins*1j]
    zi = k(np.vstack([xi.flatten(), yi.flatten()]))

    return xi, yi, zi



from pydiaid import loader
from pydiaid import method_creator
from pydiaid import method_evaluation
from pydiaid import graphs

# for data manipulation
import pandas as pd

# for creating new folders
import os

# importing for scientific and numeric manipulations
import numpy as np


def create_folder(
    paths: list,
) -> None:
    """This function creates a folder structure, where all output files are
        printed.

    Parameters:
    paths (list): a list of all path names that should be created.
    """
    for create_item in paths:
        if os.path.exists(create_item):
            next
        else:
            os.makedirs(create_item)


def library_information(
    method_conf: dict,
    save_at: str,
) -> pd.DataFrame:
    """This function exectues all steps concerning only the proteomics library:
        loading of the protemics library as a pre-filtered data frame with
        unified column names, calculating the kernel density estimation of a
        data frame, creating a density plot and histogram from the data frame,
        calculation of the charge state distribution of the precursors.

    Parameters:
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.
    save_at (str): location, where all information shouldbe stored.
    """

    library = loader.load_library(
        method_conf["input"]["library_name"],
        method_conf["input"]["analysis_software"],
        method_conf["input"]["PTM"]
        )

    # 1st calculate kernel density coordinates:
    xi, yi, zi = graphs.kernel_density_calculation(
        library,
        method_conf["graphs"]["numbins"]
        )

    # plot important plots of the library to understand the method creation
    # step
    graphs.plot_precursor_distribution_as_histogram(
        library, method_conf["graphs"],
        save_at +
        '/input_library/Histogram_precursor_distribution_in_library.pdf'
        )

    graphs.plot_density(
        xi,
        yi,
        zi,
        method_conf["graphs"],
        save_at + '/input_library/Kernel_density_distribution_library.pdf')

    # save library specific information
    dict_charge_of_precursor = method_evaluation.calculate_percentage_multiple_charged_precursors(library)
    pd.DataFrame(
        {
            "charge state of precursors": list(dict_charge_of_precursor.keys()),
            "ration [%]": list(dict_charge_of_precursor.values())
        }
        ).to_csv(
            save_at
            + '/input_library/percentage_of_multiple_charged_precursors.csv',
            index=False
            )

    return xi, yi, zi, library


def precursor_within_scan_area(
    library: pd.DataFrame,
    method_parameters: dict,
    save_at: str,
) -> None:
    """Calculation of the precursors within the scan area. The infomration
    is saved as an csv file.

    Parameters:
    library (pd.DataFrame): pre-filtered data frame with unified column
        names.
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.
    save_at (str): location, where all information should be stored.
    """

    # save parameters for method evaluation as .csv
    dict_precursors_within_scan_area = method_evaluation.calculate_precursor_within_scan_area(
        library,
        method_parameters["mz"],
        method_parameters["ion_mobility"]
    )
    print(dict_precursors_within_scan_area)

    pd.DataFrame({
        "column name": list(dict_precursors_within_scan_area.keys()),
        "column value": list(dict_precursors_within_scan_area.values())
    }).to_csv(
        save_at + '/final_method/precursors_within_scan_area.csv',
        index=False)


def create_final_method(
    library: pd.DataFrame,
    method_parameters: dict,
    dim: list,
    method_conf: dict,
) -> pd.DataFrame:
    """This function calculates the final dia-PASEF window scheme with the
        optimized scan area coordinates and writes the .txt dia-PASEF parameter
        file as input file for timsControl. By default, the scan area
        coordinates are up-shifted by 0.022 1/K0 since the quadrupole has a
        dead time at the beginning of the dia-PASEF window where no signal is
        acquired. The up-shift should prevent that this dead area lies within
        an area with a high precursor density.

    Parameters:
    library (pd.DataFrame): a pre-filtered data frame with unified column names
        containing all required precursor information.
    method_parameters (dict): dictionary, which includes all input parameters
        for creating a diaPASEF method (e.g., m/z-range, ion mobility-range,
        number of diaPASEF scans, number of ion mobility steps, the overlap of
        the diaPASEF windows, a shift of the final window scheme in the ion
        mobility axis to account for the blind spot of the changing quadrupole
        at the top of each diaPASEF window).
    dim (list): y-coordinates of the scan area = A1, A2, B1, B2. x-coordinate
        for A1 and B1 is the lower m/z-range value, x-coordinate for A2 and B2
        is the upper m/z-range value.
    method_conf (dict): this dictionary contains all input parameters for all
        sub-functions.

    Returns:
    pd.DataFrame: data frame that contains the scan type (PASEF), scan number,
        and the corresponding diaPASEF window coordinates for each window per
        scan.
    """
    file_name_method = method_conf["input"]["save_at"] + '/final_method/diaPASEF_method.txt'
    # create the diaPASEF method scheme; dim[0],dim[1],dim[2],dim[3] == A1, A2,
    # B1, B2.

    df_parameters_final = method_creator.method_creation(
        library["mz"],
        method_parameters,
        dim[0] + method_parameters["shift_of_final_method"],
        dim[1] + method_parameters["shift_of_final_method"],
        dim[2] + method_parameters["shift_of_final_method"],
        dim[3] + method_parameters["shift_of_final_method"]
    )
    method_creator.create_parameter_dataframe(
        df_parameters_final,
        file_name_method
    )
    return df_parameters_final


def final_method_information(
    df_parameters_final: pd.DataFrame,
    xi: np.ndarray,
    yi: np.ndarray,
    zi: np.ndarray,
    method_conf: dict,
    save_at: str,
    library: pd.DataFrame,
    method_parameters: dict,
    dim: list
) -> None:
    """This function calculates the evaluation parameters for the final
    dia-PASEF method and plots the dia-PASEF acquisition scheme on top of
    the kernel density pot.

    Parameters:
    df_parameters_final (pd.DataFrame): data frame that contains the scan type
        (PASEF), scan number, and the corresponding dia-PASEF window coordinates
        for each window per scan.
    xi, yi, zi (numpy.ndarray): coordinates of the kernel density estimation,
        where zi indicates the density.
    method_conf (dict): this dictionary contains all input parameters for all
            sub-functions.
    save_at (str): location, where all information shouldbe stored.
    library (pd.DataFrame): a pre-filtered data frame with unified column names
        containing all required precursor information.
    method_parameters (dict): dictionary, which includes all input parameters
        for creating a dia-PASEF method.
    dim (list): y-coordinates of the scan area = A1, A2, B1, B2. x-coordinate
        for A1 and B1 is the lower m/z-range value, x-coordinate for A2 and B2
        is the upper m/z-range value.
    """
    # plot created method on top of kernel density estimation of library
    graphs.plot_density_and_method(
        df_parameters_final,
        xi,
        yi,
        zi,
        method_conf["graphs"],
        save_at +'/final_method/Kernel_density_distribution_and_final_method.pdf'
    )
    graphs.plot_density_and_method(
        df_parameters_final,
        xi,
        yi,
        zi,
        method_conf["graphs"],
        save_at +'/final_method/Kernel_density_distribution_and_final_method.png'
    )

    # save parameters for method evaluation as .csv
    dict_evaluation_of_final_method = method_evaluation.coverage(
        df_parameters_final,
        library
    )

    if dim is None:
        next
    else:
        dict_evaluation_of_final_method["final A1, A2, B1, B2 values"] = str([
            dim[0] + method_parameters["shift_of_final_method"],
            dim[1] + method_parameters["shift_of_final_method"],
            dim[2] + method_parameters["shift_of_final_method"],
            dim[3] + method_parameters["shift_of_final_method"]
            ]
        )

    pd.DataFrame({
        "column name": list(dict_evaluation_of_final_method.keys()),
        "column value": list(dict_evaluation_of_final_method.values())
    }).to_csv(
        save_at + '/final_method/parameters_to_evaluate_method.csv',
        index=False)


def evaluate_for_multiple_charged_prec(
    method_conf: dict,
    save_at: str,
    library: pd.DataFrame,
    df_parameters_final: pd.DataFrame,
) -> None:
    """This function generates all kernel density plots with and without the
        dia-PASEF acquisition scheme on top for precursors with multiple charge
        states.

    Parameters:
    method_conf (dict): this dictionary contains all input parameters for all
            sub-functions.
    save_at (str): location, where all information should be stored.
    library (pd.DataFrame): a pre-filtered data frame with unified column names
        containing all required precursor information.
    df_parameters_final (pd.DataFrame): data frame that contains the scan type
        (PASEF), scan number, and the corresponding dia-PASEF window coordinates
        for each window per scan.
    """

    for charge in range(2, 5, 1):
        xi_charge, yi_charge, zi_charge = graphs.kernel_density_calculation_multiple_charge(
            library,
            method_conf["graphs"]["numbins"],
            charge)
        graphs.plot_density(
            xi_charge,
            yi_charge,
            zi_charge,
            method_conf["graphs"],
            save_at + '/evaluation_multiple_charged_prec/Kernel_density_distribution_library_'+ str(charge) +'x_charged_precursors.pdf')
        graphs.plot_density_and_method(
            df_parameters_final,
            xi_charge,
            yi_charge,
            zi_charge,
            method_conf["graphs"],
            save_at +'/evaluation_multiple_charged_prec/Kernel_density_distribution_and_final_method_'+ str(charge) +'x_charged_precursors.pdf'
        )
